# Copyright (c) 2021-2024  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

# Standard Library Imports
import logging
from typing import Any, Dict, Optional

# Third Party Imports

# Local Imports
from navigate.model.devices.camera.base import CameraBase
from navigate.model.devices.device_types import SequenceDevice
from navigate.model.devices.APIs.hamamatsu.HamamatsuAPI import DCAM
from navigate.tools.decorators import log_initialization

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


@log_initialization
class HamamatsuBase(CameraBase, SequenceDevice):
    """HamamatsuOrca camera class.

    This is the default parent class for Hamamatsu Cameras.
    This includes the ORCA Flash 4.0, Fusion, Lightning, and Fire.
    """

    def __init__(
        self,
        microscope_name: str,
        device_connection: Any,
        configuration: Dict[str, Any],
        *args: Optional[Any],
        **kwargs: Optional[Any],
    ) -> None:
        """Initialize HamamatsuOrca class.

        Parameters
        ----------
        microscope_name : str
            Name of microscope in configuration
        device_connection : Any
            Hardware device to connect to
        configuration : Dict[str, Any]
            Global configuration of the microscope
        """
        super().__init__(microscope_name, device_connection, configuration)

        #: str: Name of the microscope
        self.microscope_name = microscope_name

        #: object: Device connection
        self.device_connection = device_connection

        #: dict: Configuration settings
        self.configuration = configuration

        #: dict: Camera parameters
        self.camera_parameters["x_pixels"] = self.camera_controller.max_image_width
        self.camera_parameters["y_pixels"] = self.camera_controller.max_image_height
        self.camera_parameters["x_pixels_min"] = self.camera_controller.min_image_width
        self.camera_parameters["y_pixels_min"] = self.camera_controller.min_image_height
        self.camera_parameters[
            "x_pixels_step"
        ] = self.camera_controller.step_image_width
        self.camera_parameters[
            "y_pixels_step"
        ] = self.camera_controller.step_image_height
        self.minimum_exposure_time, _, _ = self.camera_controller.get_property_range(
            "exposure_time"
        )

        #: object: Camera controller
        if self.camera_controller.get_property_value("readout_speed"):
            _, speed_max, _ = self.camera_controller.get_property_range("readout_speed")
            if speed_max is not None:
                self.camera_controller.set_property_value(
                    "readout_speed", int(speed_max)
                )
                self.camera_parameters["readout_speed"] = int(speed_max)
            else:
                self.camera_controller.set_property_value("readout_speed", 1)
                self.camera_parameters["readout_speed"] = 1

        self.camera_parameters[
            "pixel_size_in_microns"
        ] = self.camera_controller.get_property_value("pixel_width")

        # Values are pulled from the CameraParameters section of the configuration.yml
        # file. Exposure time converted here from milliseconds to seconds.

        self.set_trigger_mode()

        self.camera_parameters["supported_trigger_sources"] = ["External", "Internal"]

    def __str__(self):
        """Return string representation of HamamatsuOrca class.

        Returns
        -------
        str
            String representation of HamamatsuOrca class.
        """
        return "HamamatsuBase"

    def __del__(self):
        """Delete HamamatsuOrca class."""
        self.camera_controller.dev_close()
        # self.close_camera()

    @classmethod
    def connect(cls, camera_id):
        """Connect to HamamatsuOrca camera.

        Parameters
        ----------
        camera_id : int
            Camera ID.

        Returns
        -------
        camera_controller : object
            Camera controller object.
        """
        camera_controller = DCAM(camera_id)
        return camera_controller

    @property
    def serial_number(self):
        """Get Camera Serial Number

        Returns
        -------
        serial_number : str
            Serial number for the camera.
        """
        return self.camera_controller._serial_number

    def report_settings(self):
        """Print Camera Settings.

        Prints the current camera settings to the console and the log file."""
        params = [
            "defect_correct_mode",
            "sensor_mode",
            "binning",
            "subarray_mode",
            "subarray_vsize",
            "subarray_hsize",
            "readout_speed",
            "trigger_active",
            "trigger_mode",
            "trigger_polarity",
            "trigger_source",
            "internal_line_interval",
            "internal_line_speed",
            "image_height",
            "image_width",
            "exposure_time",
            "readout_time",
        ]
        for param in params:
            print(param, self.camera_controller.get_property_value(param))
            logger.info(f"{param}, {self.camera_controller.get_property_value(param)}")

        print(
            "*** exposure time range:",
            self.camera_controller.get_property_range("exposure_time"),
        )

    def close_camera(self):
        """Close HamamatsuOrca Camera"""
        self.camera_controller.dev_close()

    def set_trigger_mode(self, trigger_source="External"):
        """Set Hamamatsu trigger source and trigger mode.
        
        Parameters
        ----------
        trigger_source : str
            Trigger source: "External" or "Internal". Default is "External".
        """
        if trigger_source == "Internal":
            # Standard trigger mode
            self.camera_controller.set_property_value("trigger_mode", 1)
            # Internal trigger.
            self.camera_controller.set_property_value("trigger_source", 1)
            logger.debug("Set camera trigger mode: Free running mode.")

        else:
            self.camera_controller.set_property_value(
                "defect_correct_mode", self.camera_parameters["defect_correct_mode"]
            )
            self.camera_controller.set_property_value(
                "trigger_active", self.camera_parameters["trigger_active"]
            )
            self.camera_controller.set_property_value(
                "trigger_mode", self.camera_parameters["trigger_mode"]
            )
            self.camera_controller.set_property_value(
                "trigger_polarity", self.camera_parameters["trigger_polarity"]
            )
            self.camera_controller.set_property_value(
                "trigger_source", self.camera_parameters["trigger_source"]
            )
            logger.debug("Set camera trigger mode: External Edge Trigger.")

    def set_sensor_mode(self, mode):
        """Set HamamatsuOrca sensor mode.

        Parameters
        ----------
        mode : str
            'Normal' or 'Light-Sheet'
        """
        modes_dict = {"Normal": 1, "Light-Sheet": 12}
        if mode in modes_dict:
            self.camera_controller.set_property_value("sensor_mode", modes_dict[mode])

            # Update minimum image sizes based on scan mode
            (
                self.camera_controller.min_image_width,
                _,
                self.camera_controller.step_image_width,
            ) = self.camera_controller.get_property_range("subarray_hsize")
            (
                self.camera_controller.min_image_height,
                _,
                self.camera_controller.step_image_height,
            ) = self.camera_controller.get_property_range("subarray_vsize")
            # update configuration dict
            self.camera_parameters[
                "x_pixels_min"
            ] = self.camera_controller.min_image_width
            self.camera_parameters[
                "y_pixels_min"
            ] = self.camera_controller.min_image_height
            self.camera_parameters[
                "x_pixels_step"
            ] = self.camera_controller.step_image_width
            self.camera_parameters[
                "y_pixels_step"
            ] = self.camera_controller.step_image_height
        else:
            print("Camera mode not supported")
            logger.debug("Camera mode not supported")

    def set_readout_direction(self, mode):
        """Set HamamatsuOrca readout direction.

        Parameters
        ----------
            mode : str
                'Top-to-Bottom', 'Bottom-to-Top', 'bytrigger', or 'diverge'.
        """
        readout_direction_dict = {
            "Top-to-Bottom": 1.0,
            "Bottom-to-Top": 2.0,
            "bytrigger": 3.0,
            "diverge": 5.0,
            "Bidirectional": 6.0,
            "Rev. Bidirectional": 7.0,
        }

        if mode in readout_direction_dict:
            self.camera_controller.set_property_value(
                "readout_direction", readout_direction_dict[mode]
            )
        else:
            print("Camera readout direction not supported")
            logger.debug("Camera readout direction not supported")

    def calculate_readout_time(self):
        """Get the duration of time needed to read out an image.

        Returns
        -------
        readout_time : float
            Duration of time needed to read out an image.
        """
        readout_time = self.camera_controller.get_property_value("readout_time")

        # with camera internal delay
        return readout_time  # + 4 * self.minimum_exposure_time

    def set_exposure_time(self, exposure_time):
        """Set HamamatsuOrca exposure time.

        Note
        ----
            Units of the Hamamatsu API are in seconds.

        Parameters
        ----------
        exposure_time : float
            Exposure time in seconds.
        """
        return self.camera_controller.set_property_value("exposure_time", exposure_time)

    def set_line_interval(self, line_interval_time):
        """Set HamamatsuOrca line interval.

        Parameters
        ----------
        line_interval_time : float
            Line interval duration.
        """
        return self.camera_controller.set_property_value(
            "internal_line_interval", line_interval_time
        )

    def get_line_interval(self):
        """Get HamamatsuOrca line interval.

        Returns
        -------
        line_interval_time : float
            Line interval duration.
        """
        self.line_interval = self.camera_controller.get_property_value(
            "internal_line_interval"
        )

    def set_binning(self, binning_string):
        """Set HamamatsuOrca binning mode.

        Parameters
        ----------
        binning_string : str
            Desired binning properties (e.g., '1x1', '2x2', '4x4', '8x8', '16x16',
            '1x2', '2x4')

        Returns
        -------
        result: bool
            True if successful, False otherwise.
        """
        binning_dict = {
            "1x1": 1,
            "2x2": 2,
            "4x4": 4,
            # '8x8': 8,
            # '16x16': 16,
            # '1x2': 102,
            # '2x4': 204
        }
        if binning_string not in binning_dict.keys():
            logger.debug(f"can't set binning to {binning_string}")
            print(f"can't set binning to {binning_string}")
            return False
        self.camera_controller.set_property_value(
            "binning", binning_dict[binning_string]
        )
        return True

    def set_ROI(self, roi_width=2048, roi_height=2048, center_x=1024, center_y=1024):
        """Change the size of the active region on the camera.

        Parameters
        ----------
        roi_width : int
            Width of active camera region.
        roi_height : int
            Height of active camera region.
        center_x : int
            X position of the center of view
        center_y : int
            Y position of the center of view

        Returns
        -------
        result: bool
            True if successful, False otherwise.
        """
        # Get the Maximum Number of Pixels from the Configuration File
        camera_height = self.camera_parameters["y_pixels"]
        camera_width = self.camera_parameters["x_pixels"]

        roi_top = center_y - roi_height // 2
        roi_bottom = center_y + roi_height // 2
        roi_left = center_x - roi_width // 2
        roi_right = center_x + roi_width // 2

        step_image_height = self.camera_controller.step_image_height
        step_image_width = self.camera_controller.step_image_width

        if (
            roi_top < 0
            or roi_bottom > camera_height
            or roi_left < 0
            or roi_right > camera_width
            or roi_top % step_image_height != 0
            or roi_bottom % step_image_height != 0
            or roi_left % step_image_width != 0
            or roi_right % step_image_width != 0
        ):
            logger.debug(f"can't set roi to {roi_width} and {roi_height}")
            return False

        # Set ROI
        self.x_pixels, self.y_pixels = self.camera_controller.set_ROI(
            roi_left, roi_top, roi_right, roi_bottom
        )

        logger.info(
            "HamamatsuOrca - subarray_hpos, "
            f"{self.camera_controller.get_property_value('subarray_hpos')}"
        )
        logger.info(
            "HamamatsuOrca - subarray_hsize, "
            f"{self.camera_controller.get_property_value('subarray_hsize')}"
        )
        logger.info(
            "HamamatsuOrca - subarray_vpos, "
            f"{self.camera_controller.get_property_value('subarray_vpos')}"
        )
        logger.info(
            "HamamatsuOrca - subarray_vsize, "
            f"{self.camera_controller.get_property_value('subarray_vsize')}"
        )

        return self.x_pixels == roi_width and self.y_pixels == roi_height

    def initialize_image_series(self, data_buffer=None, number_of_frames=100):
        """Initialize HamamatsuOrca image series.

        Parameters
        ----------
        data_buffer :
            List of SharedNDArrays of shape=(self.img_height,
            self.img_width) and dtype="uint16"
            Default is None.
        number_of_frames : int
            Number of frames.  Default is 100.
        """
        self.camera_controller.start_acquisition(data_buffer, number_of_frames)
        self.is_acquiring = True

    def close_image_series(self):
        """Close image series.

        Stops the acquisition and sets is_acquiring flag to False.
        """
        self.camera_controller.stop_acquisition()
        self.is_acquiring = False

    def get_new_frame(self):
        """Get frame from HamamatsuOrca camera.

        Returns
        -------
        frame : numpy.ndarray
            Frame ids from HamamatsuOrca camera.
        """
        return self.camera_controller.get_frames()


@log_initialization
class HamamatsuOrcaLightningCamera(HamamatsuBase):
    """HamamatsuOrcaLightning camera class."""

    def __init__(
        self,
        microscope_name: str,
        device_connection: Any,
        configuration: Dict[str, Any],
        *args: Optional[Any],
        **kwargs: Optional[Any],
    ) -> None:
        """Initialize HamamatsuOrcaLightning class.

        Parameters
        ----------
        microscope_name : str
            Name of microscope in configuration
        device_connection : Any
            Hardware device to connect to
        configuration : Dict[str, Any]
            Global configuration of the microscope
        """
        HamamatsuBase.__init__(self, microscope_name, device_connection, configuration)

        self.camera_parameters["supported_readout_directions"] = ["Top-to-Bottom"]

        # self.minimum_exposure_time = 6.304 * 10 ** -6

    def __str__(self):
        """Return string representation of HamamatsuOrcaLightning class.

        Returns
        -------
        str
            String representation of HamamatsuOrcaLightning class.
        """
        return "HamamatsuOrcaLightning"

    def calculate_light_sheet_exposure_time(
        self, full_chip_exposure_time, shutter_width
    ):
        """Calculate light sheet exposure time.

        Parameters
        ----------
        full_chip_exposure_time : float
            Full chip exposure time in seconds.
        shutter_width : int
            Shutter width.

        Returns
        -------
        exposure_time : float
            Exposure time in seconds.
        camera_line_interval : float
            Camera line interval in seconds.
        full_chip_exposure_time : float
            Full chip exposure time in seconds.
        """

        camera_line_interval = full_chip_exposure_time / (
            6 + (shutter_width + self.y_pixels) / 4
        )
        self.camera_parameters["line_interval"] = camera_line_interval

        maximum_internal_line_interval = 0.0002  # 200.0 us
        if camera_line_interval > maximum_internal_line_interval:
            camera_line_interval = maximum_internal_line_interval
            full_chip_exposure_time = camera_line_interval * (
                6 + (shutter_width + self.y_pixels) / 4
            )

        exposure_time = camera_line_interval * ((shutter_width + 3) // 4)
        return exposure_time, camera_line_interval, full_chip_exposure_time


@log_initialization
class HamamatsuOrcaFireCamera(HamamatsuBase):
    def __init__(
        self,
        microscope_name: str,
        device_connection: Any,
        configuration: Dict[str, Any],
        *args: Optional[Any],
        **kwargs: Optional[Any],
    ) -> None:
        """Initialize HamamatsuOrcaFire class.

        Parameters
        ----------
        microscope_name : str
            Name of microscope in configuration
        device_connection : Any
            Hardware device to connect to
        configuration : Dict[str, Any]
            Global configuration of the microscope
        """
        HamamatsuBase.__init__(self, microscope_name, device_connection, configuration)
        self.camera_parameters["supported_readout_directions"] = [
            "Top-to-Bottom",
            "Bottom-to-Top",
            "Bidirectional",
            "Rev. Bidirectional",
        ]

        # self.minimum_exposure_time = 7.309 * 10 ** -6  # 7.309 us

        self.camera_parameters["x_pixels"] = self.camera_controller.get_property_value(
            "image_width"
        )

        self.camera_parameters["y_pixels"] = self.camera_controller.get_property_value(
            "image_height"
        )

    def __str__(self):
        """Return string representation of HamamatsuOrcaFire class.

        Returns
        -------
        str
            String representation of HamamatsuOrcaFire class.
        """
        return "HamamatsuOrcaFire"

    def calculate_light_sheet_exposure_time(
        self, full_chip_exposure_time, shutter_width
    ):
        """Convert normal mode exposure time to light-sheet mode exposure time.
        Calculate the parameters for an acquisition

        Parameters
        ----------
        full_chip_exposure_time : float
            Normal mode exposure time in seconds.
        shutter_width : int
            Width of light-sheet rolling shutter.

        Returns
        -------
        exposure_time : float
            Light-sheet mode exposure time (s).
        camera_line_interval : float
            HamamatsuOrca line interval duration (s).
        full_chip_exposure_time : float
            Updated full chip exposure time (s).
        """
        # 4H delay, (Vn/2+5)H readout
        camera_line_interval = full_chip_exposure_time / (
            9 + (shutter_width + self.y_pixels) / 2
        )

        maximum_internal_line_interval = 0.0002339  # 233.9 us
        if camera_line_interval > maximum_internal_line_interval:
            camera_line_interval = maximum_internal_line_interval
            full_chip_exposure_time = camera_line_interval * (
                9 + (shutter_width + self.y_pixels) / 2
            )

        self.camera_parameters["line_interval"] = camera_line_interval

        # round up exposure time
        exposure_time = camera_line_interval * ((shutter_width + 1) // 2)
        return exposure_time, camera_line_interval, full_chip_exposure_time


@log_initialization
class HamamatsuOrcaCamera(HamamatsuBase):
    def __init__(
        self,
        microscope_name: str,
        device_connection: Any,
        configuration: Dict[str, Any],
        *args: Optional[Any],
        **kwargs: Optional[Any],
    ) -> None:
        """Initialize HamamatsuOrca class.

        This is for controlling the Orca Flash 4.0.

        Parameters
        ----------
        microscope_name : str
            Name of microscope in configuration
        device_connection : Any
            Hardware device to connect to
        configuration : Dict[str, Any]
            Global configuration of the microscope
        """
        HamamatsuBase.__init__(self, microscope_name, device_connection, configuration)

        self.camera_parameters["supported_readout_directions"] = [
            "Top-to-Bottom",
            "Bottom-to-Top",
        ]

        # self.minimum_exposure_time = 9.74436 * 10 ** -6

    def __str__(self):
        """Return string representation of HamamatsuOrca class.

        Returns
        -------
        str
            String representation of HamamatsuOrca class.
        """
        return "HamamatsuOrca"

    def calculate_light_sheet_exposure_time(
        self, full_chip_exposure_time, shutter_width
    ):
        """Convert normal mode exposure time to light-sheet mode exposure time.
        Calculate the parameters for an acquisition

        Parameters
        ----------
        full_chip_exposure_time : float
            Normal mode exposure time in seconds.
        shutter_width : int
            Width of light-sheet rolling shutter.

        Returns
        -------
        exposure_time : float
            Light-sheet mode exposure time (s).
        camera_line_interval : float
            HamamatsuOrca line interval duration (s).
        full_chip_exposure_time : float
            Updated full chip exposure time (s).
        """
        camera_line_interval = full_chip_exposure_time / (
            10 + shutter_width + self.y_pixels
        )

        maximum_internal_line_interval = 0.1  # 100ms
        if camera_line_interval > maximum_internal_line_interval:
            camera_line_interval = maximum_internal_line_interval
            full_chip_exposure_time = camera_line_interval * (
                10 + shutter_width + self.y_pixels
            )

        self.camera_parameters["line_interval"] = camera_line_interval * shutter_width

        # round up exposure time
        exposure_time = camera_line_interval * shutter_width
        return exposure_time, camera_line_interval, full_chip_exposure_time


@log_initialization
class HamamatsuOrcaFusionCamera(HamamatsuBase):
    """HamamatsuOrcaFusion camera class."""

    def __init__(
        self,
        microscope_name: str,
        device_connection: Any,
        configuration: Dict[str, Any],
        *args: Optional[Any],
        **kwargs: Optional[Any],
    ) -> None:
        """Initialize HamamatsuOrcaFusion class.

        Parameters
        ----------
        microscope_name : str
            Name of microscope in configuration
        device_connection : Any
            Hardware device to connect to
        configuration : Dict[str, Any]
            Global configuration of the microscope
        """
        HamamatsuBase.__init__(self, microscope_name, device_connection, configuration)

        self.camera_parameters["supported_readout_directions"] = [
            "Top-to-Bottom",
            "Bottom-to-Top",
        ]

    def __str__(self):
        """Return string representation of HamamatsuOrcaFusion class.

        Returns
        -------
        str
            String representation of HamamatsuOrcaFusion class.
        """
        return "HamamatsuOrcaFusion"

    def calculate_light_sheet_exposure_time(
        self, full_chip_exposure_time, shutter_width
    ):
        """Calculate light sheet exposure time.

        Parameters
        ----------
        full_chip_exposure_time : float
            Full chip exposure time in seconds.
        shutter_width : int
            Shutter width.

        Returns
        -------
        exposure_time : float
            Exposure time in seconds.
        camera_line_interval : float
            Camera line interval in seconds.
        full_chip_exposure_time : float
            Full chip exposure time in seconds.
        """

        camera_line_interval = full_chip_exposure_time / (
            4 + shutter_width + self.y_pixels
        )

        maximum_internal_line_interval = 963.8e-6  # 963.8 us
        if camera_line_interval > maximum_internal_line_interval:
            camera_line_interval = maximum_internal_line_interval
            full_chip_exposure_time = camera_line_interval * (
                4 + shutter_width + self.y_pixels
            )

        self.camera_parameters["line_interval"] = camera_line_interval * shutter_width

        # round up exposure time
        exposure_time = camera_line_interval * shutter_width
        return exposure_time, camera_line_interval, full_chip_exposure_time



# Copyright (c) 2021-2024  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Standard Library Imports
import platform
import threading
import tkinter as tk
from typing import Any
import time

# Third Party Imports
import numpy as np
from matplotlib.ticker import FuncFormatter

from navigate.config import update_config_dict

# Local Imports
from navigate.model.concurrency.concurrency_tools import SharedNDArray
from navigate.view.main_window_content.display_notebook import HistogramFrame


# Logger Setup
# p = __name__.split(".")[1]
# logger = logging.getLogger(p)


class HistogramController:
    """Histogram controller"""

    def __init__(self, histogram: HistogramFrame, parent_controller: Any) -> None:
        """Initialize the histogram controller

        Parameters
        ----------
        histogram : HistogramFrame
            Histogram view
        parent_controller : Any
            Main controller.
        """

        #: HistogramFrame: Histogram view
        self.histogram = histogram

        #: MainController: Main controller
        self.parent_controller = parent_controller

        #: FigureBase: The histogram figure.
        self.ax = self.histogram.figure.add_axes([0.075, 0.25, 0.88, 0.65])
        self.ax.tick_params(
            axis="both", which="both", direction="inout", labelsize=8, reset=True
        )

        # Event Bindings
        widget = self.histogram.figure_canvas.get_tk_widget()

        if platform.system() == "Darwin":
            widget.bind("<Button-2>", self.histogram_popup)
        else:
            widget.bind("<Button-3>", self.histogram_popup)

        # Default axis values
        self.x_axis_var = tk.StringVar(value="linear")
        self.y_axis_var = tk.StringVar(value="log")

        #: bool: Histogram enabled
        self.histogram_enabled = tk.BooleanVar()

        #: bool: Logarithmic X-axis
        self.log_x = False

        #: bool: Logarithmic Y-axis
        self.log_y = True

        #: tk.Menu: Histogram popup menu
        self.menu = tk.Menu(widget, tearoff=0)
        self.menu.add_radiobutton(
            label="Log X",
            variable=self.x_axis_var,
            value="log",
            command=self.update_scale,
        )
        self.menu.add_radiobutton(
            label="Linear X",
            variable=self.x_axis_var,
            value="linear",
            command=self.update_scale,
        )
        self.menu.add_separator()
        self.menu.add_radiobutton(
            label="Log Y",
            variable=self.y_axis_var,
            value="log",
            command=self.update_scale,
        )
        self.menu.add_radiobutton(
            label="Linear Y",
            variable=self.y_axis_var,
            value="linear",
            command=self.update_scale,
        )
        self.menu.add_separator()
        self.menu.add_checkbutton(
            label="Enable Histogram",
            variable=self.histogram_enabled,
            onvalue=True,
            offvalue=False,
            command=self.update_experiment,
        )

        #: threading.Thread: Histogram thread
        self.histogram_thread = None

        #: threading.Lock: Lock
        self.lock = threading.Lock()

        # Default location for communicating with the plugin in the model.
        if "histogram" not in self.parent_controller.configuration["gui"].keys():
            update_config_dict(
                manager=self.parent_controller.manager,
                parent_dict=self.parent_controller.configuration["gui"],
                config_name="histogram",
                new_config={"enabled": True},
            )

        # Set histogram according to the experiment.yaml file. If disabled, stays disabled upon restart.
        self.histogram_enabled.set(
            self.parent_controller.configuration["gui"]["histogram"].get(
                "enabled", True
            )
        )

    def update_experiment(self) -> None:
        """Update the experiment.yaml file. Also communicate any changes to the menu
        controller."""
        # Get the state of the histogram enabled variable.
        histogram_state = self.histogram_enabled.get()

        # Update the experiment.yaml file.
        self.parent_controller.configuration["gui"]["histogram"][
            "enabled"
        ] = histogram_state

        # Communicate changes to the menu controller.
        self.parent_controller.menu_controller.histogram_enabled.set(histogram_state)

    def update_scale(self) -> None:
        """Update the scale of the histogram"""
        self.log_x = self.x_axis_var.get() == "log"
        self.log_y = self.y_axis_var.get() == "log"

    def histogram_popup(self, event: tk.Event) -> None:
        """Histogram popup menu

        Parameters
        ----------
        event : tk.Event
            Event
        """
        try:
            self.menu.tk_popup(event.x_root, event.y_root, 0)
        finally:
            self.menu.grab_release()

    def populate_histogram(self, image: SharedNDArray) -> None:
        """Populate the histogram in a dedicated thread.

        Parameters
        ----------
        image : SharedNDArray
            Image data
        """
        if not self.histogram_enabled.get():
            # Draw a disabled state
            if self.histogram_thread is None or not self.histogram_thread.is_alive():
                self.histogram_thread = threading.Thread(target=self._clear_histogram)
                self.histogram_thread.start()
            return

        if self.histogram_thread and self.histogram_thread.is_alive():
            return

        self.histogram_thread = threading.Thread(
            target=self._populate_histogram, args=(image,)
        )
        self.histogram_thread.start()

    def _populate_histogram(self, image: SharedNDArray) -> None:
        """Populate the histogram

        We continue to reduce the size of the histogram to make it more efficient.
        To estimate a distribution within ±ε accuracy in total variation distance (
        TVD), you typically need O(B / ε²) samples, where B is the number of
        histogram bins, and ε is the desired accuracy.

        Parameters
        ----------
        image : SharedNDArray
            Image Data
        """

        # Estimate total variation distance.
        number_bins = 20
        accuracy = 0.05  # 5% accuracy in histogram.
        required_pixels = number_bins / (accuracy**2)
        actual_pixels = image.size

        # Downsample data to the required number of pixels to meet accuracy.
        down_sampling_constant = max(1, int(actual_pixels // required_pixels))
        data = image.flatten()
        data = data[::down_sampling_constant]

        # Plot histogram.
        self.ax.cla()
        counts, bins = np.histogram(data, bins=number_bins)
        self.ax.bar(bins[:-1], counts, width=np.diff(bins), color="black", align="edge")

        x_maximum = np.max(data) + np.std(data)
        x_minimum = np.min(data) - np.std(data)
        x_minimum = 1 if x_minimum < 1 else x_minimum
        y_maximum = 10**6 // down_sampling_constant

        self.ax.set_xlim(x_minimum, x_maximum)
        self.ax.set_ylim(1, y_maximum)

        if self.log_y:
            self.ax.set_yscale("log")

        if self.log_x:
            self.ax.set_xscale("log")

        self.ax.yaxis.set_major_formatter(
            FuncFormatter(
                lambda val, pos: f"$10^{{{int(np.log10(val))}}}$" if val > 0 else ""
            )
        )
        self.histogram.figure_canvas.draw()

    def _clear_histogram(self) -> None:
        """Clear the histogram but keep canvas interactive."""
        self.ax.cla()
        self.ax.text(
            x=0.5,
            y=0.5,
            s="Intensity Histogram Disabled\nRight Click to Enable",
            fontdict={
                "family": "Arial",
                "size": 10,
                "style": "italic",
                "color": "gray",
            },
            ha="center",
            va="center",
            bbox=dict(facecolor="white", edgecolor="none", boxstyle="round,pad=0.5"),
            transform=self.ax.transAxes,
        )
        self.ax.set_xticks([])
        self.ax.set_yticks([])
        self.histogram.figure_canvas.draw()



# Copyright (c) 2021-2024  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Standard Library Imports
import platform
import logging
import time
import importlib
from multiprocessing.managers import ListProxy
from typing import Callable, Tuple, Any, Type, Dict, Optional

# Third Party Imports

# Local Imports
from navigate.tools.common_functions import build_ref_name, load_param_from_module
from navigate.model.devices.device_types import (
    SerialDevice,
    IntegratedDevice,
    NIDevice,
    SequenceDevice,
)
from navigate.model.devices.daq.base import DAQBase

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


class DummyDeviceConnection:
    """Dummy Device"""

    pass


def auto_redial(
    func: Callable[..., Any],
    args: Tuple[Any, ...],
    n_tries: int = 10,
    exception: Type[Exception] = Exception,
    **kwargs: Any,
) -> Any:
    """Retries connections to a startup device defined by `func` for a specified
    number of attempts.

    This function attempts to execute the connection function `func` up to `n_tries`
    times. If an exception occurs, it retries the connection after a brief pause,
    logging each failure. If the connection partially succeeds, it cleans up any objects
    before retrying.

    Parameters
    ----------
    func : Callable[..., Any]
        The function or class (`__init__()` method) that connects to a device.
    args : Tuple[Any, ...]
        Positional arguments to pass to the `func`.
    n_tries : int
        The number of attempts to retry the connection. Default is 10.
    exception : Type[Exception]
        The exception type to catch and handle during connection attempts.
        Default is `Exception`.
    **kwargs : Any
        Additional keyword arguments passed to `func`.

    Returns
    -------
    Any
        The result of the successful execution of `func`.

    Raises
    ------
    exception
        If all connection attempts fail, the specified `exception` is raised.
    """
    val = None

    for i in range(n_tries):
        try:
            val = func(*args, **kwargs)
        except exception as e:
            if i < (n_tries - 1):
                logger.debug(
                    f"auto_redial - Failed {str(func)} attempt {i+1}/{n_tries} "
                    f"with exception {e}."
                )
                # If we failed, but part way through object creation, we must
                # delete the object prior to trying again. This lets us restart
                # the connection process with a clean slate
                if val is not None:
                    val.__del__()
                    del val
                    val = None
                time.sleep(0.5)  # TODO: 0.5 reached by trial and error. Better value?
            else:
                logger.error(f"Device startup failed: {e}")
                raise exception
        else:
            break

    return val


class SerialConnectionFactory:
    """Serial Connection Factory.

    This class is used to build serial connections to devices.
    """

    _connections = {}

    @classmethod
    def build_connection(
        cls,
        build_connection_function: Callable[..., Any],
        args: Tuple[Any, ...],
        exception: Type[Exception] = Exception,
    ) -> Any:
        """
        Builds a serial connection to a device.

        This method establishes a connection to a device using the provided
        connection-building function and arguments. If the connection does not exist,
        it will be created and stored.

        Parameters
        ----------
        build_connection_function : Callable
            Function that builds the connection to the device.
        args : Tuple
            Arguments to the build_connection_function
        exception : Type[Exception]
            Exception to catch when building the connection

        Returns
        -------
        connection : Any
            Connection to the device

        Raises
        ------
        exception : Exception
            If the connection building process fails, the specified `exception` is
            raised.
        """
        port = args[0]
        if str(port) not in cls._connections:
            cls._connections[str(port)] = auto_redial(
                build_connection_function, args, exception=exception
            )

        return cls._connections[str(port)]


class IntegratedDeviceFactory:
    """Integrated Device Factory.

    This class is used to build integrated device connections.
    """

    _connections = {}

    @classmethod
    def build_connection(
        cls,
        unique_id: str,
        build_connection_function: Callable[..., Any],
        args: Tuple[Any, ...],
        exception: Type[Exception] = Exception,
    ) -> Any:
        """
        Builds an integrated device connection.

        This method establishes a connection to an integrated device using the provided
        connection-building function and arguments. If the connection does not exist, it
        will be created and stored.

        Parameters
        ----------
        unique_id : str
            Unique identifier for the device
        build_connection_function : Callable
            Function that builds the connection to the device.
        args : Tuple
            Arguments to the build_connection_function
        exception : Type[Exception]
            Exception to catch when building the connection

        Returns
        -------
        connection : Any
            Connection to the device

        Raises
        ------
        exception : Exception
            If the connection building process fails, the specified `exception` is
            raised.
        """
        if str(unique_id) not in cls._connections:
            cls._connections[str(unique_id)] = auto_redial(
                build_connection_function, args, exception=exception
            )

        return cls._connections[str(unique_id)]


class SequenceDeviceFactory:
    """Sequence Device Factory.

    This class is used to build sequence device connections.
    """

    _connections = {}

    @classmethod
    def build_connection(
        cls,
        unique_id: str,
        build_connection_function: Callable[..., Any],
        args: Tuple[Any, ...],
        exception: Type[Exception] = Exception,
        max_device_num: int = 2,
    ) -> Any:
        """
        Builds a sequence device connection.

        This method establishes a connection to a sequence device using the provided
        connection-building function and arguments. If the connection does not exist, it
        will be created and stored.

        Parameters
        ----------
        unique_id : str
            Unique identifier for the device
        build_connection_function : Callable
            Function that builds the connection to the device.
        args : Tuple
            Arguments to the build_connection_function
        exception : Type[Exception]
            Exception to catch when building the connection
        max_device_num : int
            Maximum number of devices to connect to. Default is 2.

        Returns
        -------
        connection : Any
            Connection to the device

        Raises
        ------
        exception : Exception
            If the connection building process fails, the specified `exception` is
            raised.
        """
        device_type, serial_number = unique_id.split("_")
        if device_type not in cls._connections:
            cls._connections[device_type] = {}

        if str(serial_number) in cls._connections[device_type]:
            return cls._connections[device_type][str(serial_number)]

        for i in range(len(cls._connections[device_type].keys()), max_device_num):
            device = auto_redial(build_connection_function, (i,), exception=exception)
            cls._connections[device_type][device._serial_number] = device
            if device._serial_number[0] == "0":
                try:
                    oct_num = int(device._serial_number, 8)
                    cls._connections[device_type][str(oct_num)] = device
                except ValueError:
                    logger.debug("Error converting device serial number to octal")
                    pass
            if str(serial_number) in cls._connections[device_type]:
                return cls._connections[device_type][str(serial_number)]

        raise device_not_found(device_type, serial_number)


def start_device(
    microscope_name: str,
    configuration: Dict[str, Any],
    device_category: str,
    device_id: int = 0,
    is_synthetic: bool = False,
    daq_connection: Optional[Any] = None,
    plugin_devices: Optional[Dict] = None,
) -> Any:
    """Starts a device.

    Parameters
    ----------
    microscope_name : str
        Name of microscope in configuration
    configuration : Dict[str, Any]
        Global configuration of the microscope
    device_category : str
        Type of device to connect to, such as stage, camera, filter_wheel, etc.
    device_id : int
        Index of device in the configuration dictionary. Default is 0.
    is_synthetic : bool
        Run synthetic version of hardware. Default is False.
    daq_connection: Optional[Any]
        Data acquisition connection. Default is None.
    plugin_devices : Optional[Dict]
        Dictionary of plugin devices. Default is None.

    Returns
    -------
    device : Any
        Device class.
    """
    if plugin_devices is None:
        plugin_devices = {}

    device_config = configuration["configuration"]["microscopes"][microscope_name][
        device_category
    ]

    if device_category == "stage":
        device_config = device_config["hardware"]

    if type(device_config) == ListProxy:
        device_config = device_config[device_id]

    if device_category != "stage":
        device_config = device_config["hardware"]

    device_type = device_config["type"]

    if "_" in device_category:
        class_name_suffix = "".join(
            [x.capitalize() for x in device_category.split("_")]
        )
    else:
        class_name_suffix = device_category.capitalize()

    device_types_dict = load_param_from_module(
        "navigate.config.configuration_database", device_category + "_device_types"
    )

    for k in device_types_dict:
        if type(device_types_dict[k]) is not tuple:
            device_types_dict[k] = (device_types_dict[k], device_types_dict[k])

    temp_device_ref = dict(device_types_dict.values())

    if is_synthetic or device_type.lower().startswith("synthetic"):
        device_type = "Synthetic"
    elif device_type.endswith(class_name_suffix):
        device_type = device_type[: -len(class_name_suffix)]

    # device type naming rules: manufacturer(file_name).device_model
    # if the device is currently supported in Navigate, you can use device_model as the
    # device type.
    if "." in device_type:
        device_manufacturer, device_type = device_type.split(".")[:2]
    elif device_type in temp_device_ref:
        device_manufacturer = temp_device_ref[device_type]
    else:
        device_manufacturer = None

    if device_manufacturer is not None and importlib.util.find_spec(
        f"navigate.model.devices.{device_category}.{device_manufacturer}"
    ):
        module = importlib.import_module(
            f"navigate.model.devices.{device_category}.{device_manufacturer}"
        )
        try:
            _class = getattr(module, device_type + class_name_suffix)
        except (KeyError, AttributeError):
            logger.error(
                f"There is no device class {device_type + class_name_suffix} in the file: {device_manufacturer}.py"
            )
            return device_not_found(
                microscope_name, device_category, device_type, device_id
            )

        # Load serial devices via the SerialDevice factory.
        if issubclass(_class, SerialDevice):
            device_connection = SerialConnectionFactory.build_connection(
                _class.connect,
                (
                    device_config["port"],
                    device_config.get("baudrate", 115200),
                    device_config.get("timeout", 0.25),
                ),
                exception=Exception,
            )

        # Load integrated devices via the IntegratedDevice factory.
        elif issubclass(_class, IntegratedDevice):
            # get connection parameters
            connection_params = []
            for param in _class.get_connect_params():
                connection_params.append(device_config[param])
            # build connection
            device_connection = IntegratedDeviceFactory.build_connection(
                build_ref_name(
                    "_",
                    device_manufacturer,
                    device_type,
                    device_config.get("serial_number", "000000"),
                ),
                _class.connect,
                connection_params,
                exception=Exception,
            )
        elif issubclass(_class, SequenceDevice):
            # get connection parameters
            connection_params = []
            if hasattr(_class, "get_connect_params"):
                for param in _class.get_connect_params():
                    connection_params.append(device_config[param])
            # build connection
            device_connection = SequenceDeviceFactory.build_connection(
                build_ref_name(
                    "_", device_category, device_config.get("serial_number", "000000")
                ),
                _class.connect,
                args=connection_params,
                exception=Exception,
            )
        elif hasattr(_class, "connect"):
            # get connection parameters
            connection_params = []
            if hasattr(_class, "get_connect_params"):
                for param in _class.get_connect_params():
                    connection_params.append(device_config[param])
            device_connection = auto_redial(
                _class.connect, connection_params, exception=Exception
            )
        else:
            device_connection = None

        if issubclass(_class, NIDevice):
            if device_connection is None:
                device_connection = daq_connection
            else:
                device_connection = {
                    "connection": device_connection,
                    "daq_connection": daq_connection,
                }

        return _class(microscope_name, device_connection, configuration, device_id)

    elif device_category in plugin_devices:
        # device_category in ["stage", "shutter", "filter_wheel", "remote_focus", "camera", "galvo", "zoom", "laser"]
        if device_category == "stage":
            hardware_configuration = configuration["configuration"]["microscopes"][
                microscope_name
            ][device_category]["hardware"][device_id]
        else:
            hardware_configuration = configuration["configuration"]["microscopes"][
                microscope_name
            ][device_category][device_id]["hardware"]
        # find the device in plugins
        if device_type + class_name_suffix in plugin_devices[device_category]:
            device_type += class_name_suffix
        elif device_type not in plugin_devices[device_category]:
            device_not_found(microscope_name, device_category, device_type, device_id)

        try:

            device_connection = plugin_devices[device_category][device_type][
                "load_device"
            ](
                hardware_configuration,
                is_synthetic,
                device_type=device_category,
            )
            start_function = plugin_devices[device_category][device_type][
                "start_device"
            ]
            return start_function(
                microscope_name,
                device_connection,
                configuration,
                is_synthetic=is_synthetic,
                id=device_id,
                device_type=device_category,
            )
        except RuntimeError:
            print(
                f"{device_category}:{device_type} can't be loaded correctly, "
                f"make sure you have specify SUPPORTED_DEVICE_TYPES in your plugin!"
            )
    device_not_found(microscope_name, device_category, device_type, device_id)


def start_daq(
    configuration: Dict[str, Any], device_type: str = "NI", name: str = "name"
) -> DAQBase:
    """Initializes the data acquisition (DAQ) class on a dedicated thread.

    Load daq information from the configuration file. Proper daq types include NI and
    SyntheticDAQ.

    Parameters
    ----------
    configuration : Dict[str, Any]
        Global configuration
    device_type : str
        Type of device to connect to. Default is "NI".

    Returns
    -------
    DAQ : DAQBase
        DAQ class.
    """

    if device_type == "NI":
        from navigate.model.devices.daq.ni import NIDAQ

        return NIDAQ(configuration)

    elif device_type == "asi.ASI":
        # from navigate.model.devices.daq.asi import ASIDaq
        # name = "Microscope-0"
        return start_device(name, configuration, "daq")

    elif device_type.lower().startswith("synthetic"):
        from navigate.model.devices.daq.synthetic import SyntheticDAQ

        return SyntheticDAQ(configuration)

    else:
        device_not_found("DAQ", device_type)


def device_not_found(*args: Any) -> None:
    """
    Display an error message and raise an exception if the device is not found.

    This function logs an error message and raises a `RuntimeError` when the
    specified device cannot be found in the configuration or the configuration is
    incorrect.

    Parameters
    ----------
    args : tuple
        A list of arguments representing the device details. These are typically:

        - microscope_name (str): The name of the microscope.
        - device (str): The name of the device.
        - device id (int): The ID of the device.
        - device type (str): The type of the device.

    Raises
    ------
    RuntimeError
        If the device cannot be found or the configuration is incorrect.
    """
    error_statement = f"Device not found in configuration: {args}"
    logger.error(error_statement)
    print(error_statement)
    raise RuntimeError()


def load_devices(
    microscope_name: str,
    configuration: Dict[str, Any],
    is_synthetic: bool = False,
    devices_dict: dict = {},
    plugin_devices: Optional[dict] = None,
) -> dict:
    """Load devices from configuration.

    Parameters
    ----------
    microscope_name : str
        Name of microscope in configuration
    configuration : Dict[str, Any]
        Configuration dictionary
    is_synthetic : bool
        Run synthetic version of hardware?
    devices_dict : dict
        Dictionary of devices to load
    plugin_devices : dict
        Dictionary of plugin devices

    Returns
    -------
    devices : dict
        Dictionary of devices
    """

    if plugin_devices is None:
        plugin_devices = {}

    # start daq
    if "daq" not in devices_dict:
        devices_dict["daq"] = {}
    if is_synthetic:
        device_type = "Synthetic"
    else:
        device_type = configuration["configuration"]["microscopes"][microscope_name][
            "daq"
        ]["hardware"]["type"]

    if device_type not in devices_dict["daq"]:
        devices_dict["daq"][device_type] = start_daq(configuration, device_type, microscope_name)

    return devices_dict



# Copyright (c) 2021-2024  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Standard library imports
import math
from typing import Optional
from itertools import product

# Third party imports
from skimage import filters
from skimage.transform import downscale_local_mean
from skimage import measure
from scipy.ndimage import median_filter, binary_fill_holes

import numpy as np
import numpy.typing as npt

# Local application imports
# from navigate.model.analysis.camera import compute_signal_to_noise


def has_tissue(
    image_data: npt.ArrayLike,
    x: int,
    y: int,
    width: int,
    offset: Optional[npt.ArrayLike] = None,
    variance: Optional[npt.ArrayLike] = None,
) -> bool:
    """
    Determine if an image contains tissue.

    Parameters
    ----------
    image_data : npt.ArrayLike
        Image
    x : int
        Starting x-position of subimage. Must be smaller than image_data.shape[0].
    y : int
        Starting y-position of subimage. Must be smaller than image_data.shape[1].
    width : int
        Width of subimage. Must be smaller than min(image_data.shape[:1])
    offset : npt.ArrayLike
        Camera pixel offset map. Same size as image_data.
    variance : npt.ArrayLike
        Camera pixel variance map. Same size as image_data.

    Returns
    -------
    bool
        Is tissue present?
    """
    # TODO: threshold value, other is_tissue() method
    xsl = slice(x * width, (x + 1) * width)
    ysl = slice(y * width, (y + 1) * width)

    # Decide tissue is present by hard SNR threshold
    # if offset is not None and variance is not None:
    #     threshold_value = 2  # snr threshold
    #     snr = compute_signal_to_noise(
    #         image_data[xsl, ysl], offset[xsl, ysl], variance[xsl, ysl]
    #     )
    #     return (
    #         np.sum(snr > threshold_value) / np.prod(snr.shape) > 0.0625
    #     )  # at least 6.25% of pixels have to be above this value

    # Decide tissue is present by hard pixel count threshold
    # threshold_value = 1000
    return np.any(image_data[xsl, ysl] > np.mean(image_data))


def find_tissue_boundary_2d(
    image_data: npt.ArrayLike, mag_ratio: Optional[float] = 1.0
) -> list:
    """
    Find all pixels containing tissue, based on an Otsu threshold. Optionally,
    return pixels in image space resampled by mag_ratio.

    Parameters
    ----------
    image_data : npt.ArrayLike
        Image
    mag_ratio : float
        Ratio between pixel sizes of current over target tiles.

    Returns
    -------
    boundary : list
        List of boundaries of tissue by row of downsampled image.
    """

    # Threshold
    thresh_img = image_data > filters.threshold_otsu(image_data)

    if mag_ratio > 1:
        ds_img = downscale_local_mean(thresh_img, (mag_ratio, mag_ratio))
    else:
        ds_img = image_data
        mag_ratio = 1

    idx_x, idx_y = np.where(ds_img)

    # Assume square image
    m = math.ceil(image_data.shape[0] / mag_ratio)
    # n = math.ceil(image_data.shape[1] / mag_ratio)
    boundary = [None] * m
    for x, y in zip(idx_x, idx_y):
        if boundary[x] is None:
            boundary[x] = [y, y]
        else:
            boundary[x][1] = y
    return boundary


def binary_detect(
    img_data: npt.ArrayLike,
    boundary: list,
    width: int = 1,
    offset: Optional[npt.ArrayLike] = None,
    variance: Optional[npt.ArrayLike] = None,
):
    """
    Perform a binary search for tissue on an image. Return locations of pixels
    containing tissue.

    Parameters
    ----------
    img_data : npt.ArrayLike
        Image
    boundary : dict
        List of xy pixel positions indicating presence of tissue as values.
    width : int
        Width of subimage. Must be smaller than min(image_data.shape[:1])
    offset : npt.ArrayLike
        Camera pixel offset map. Same size as image_data.
    variance : npt.ArrayLike
        Camera pixel variance map. Same size as image_data.

    Returns
    -------
    boundary : list
        List of boundaries of tissue by row of downsampled image.
    """
    m, n = img_data.shape
    m = int(m / width)
    n = int(n / width)

    def binary_search_func_left(row, left, right):
        """Binary search function.

        Parameters
        ----------
        row : int
            Row index of image.
        left : int
            Leftmost column index of subimage.
        right : int
            Rightmost column index of subimage.

        Returns
        -------
        int
            Leftmost column index of subimage.
        """
        while left < right:
            mid = (left + right) // 2
            if has_tissue(img_data, row, mid, width, offset, variance):
                right = mid
            else:
                left = mid + 1
        return right

    def binary_search_func_right(row, left, right):
        """Binary search function.

        Parameters
        ----------
        row : int
            Row index of image.
        left : int
            Leftmost column index of subimage.
        right : int
            Rightmost column index of subimage.

        Returns
        -------
        int
            Rightmost column index of subimage.
        """

        while left < right:
            mid = (left + right) // 2
            if has_tissue(img_data, row, mid, width, offset, variance):
                left = mid + 1
            else:
                right = mid
        return right - 1

    def find_tissue_range(row, left, right):
        """Find range of tissue in image.

        Parameters
        ----------
        row : int
            Row index of image.
        left : int
            Leftmost column index of subimage.
        right : int
            Rightmost column index of subimage.

        Returns
        -------
        int
            Leftmost column index of subimage.
        int
            Middle column index of subimage.
        int
            Rightmost column index of subimage.
        """

        temp = [(left, right)]
        while temp:
            temp2 = []
            for ll, r in temp:
                mid = (ll + r) // 2
                if has_tissue(img_data, row, mid, width, offset, variance):
                    return ll, mid, r
                if mid > ll + 1:
                    temp2.append((ll, mid))
                if r > mid + 1:
                    temp2.append((mid, r))
            temp = temp2
        return -1, -1, -1

    def detect_row_boundary(row_id, left, right):
        """Detect row boundary.

        Parameters
        ----------
        row_id : int
            Row index of image.
        left : int
            Leftmost column index of subimage.
        right : int
            Rightmost column index of subimage.

        Returns
        -------
        int
            Leftmost column index of subimage.
        int
            Rightmost column index of subimage.
        """
        is_tissue_left = has_tissue(img_data, row_id, left, width, offset, variance)
        is_tissue_right = has_tissue(img_data, row_id, right, width, offset, variance)

        if is_tissue_left and is_tissue_right:
            left_l, left_r = 0, left
            right_l, right_r = right, n
        elif is_tissue_left:
            left_l, left_r = 0, left
            right_l, right_r = left, right
        elif is_tissue_right:
            left_l, left_r = left, right
            right_l, right_r = right, n
        else:
            ll, mid, r = find_tissue_range(row_id, left, right)
            left_l, left_r = ll, mid
            right_l, right_r = mid, r

        if left_l == -1:
            return None, None
        return binary_search_func_left(
            row_id, left_l, left_r
        ), binary_search_func_right(row_id, right_l, right_r)

    def expand_row(row_id, limits, direction, boundary):
        """Expand row.

        Parameters
        ----------
        row_id : int
            Row index of image.
        limits : int
            Limits of row.
        direction : int
            Direction of row.
        boundary : list
            List of boundaries of tissue by row of downsampled image.

        Returns
        -------
        new_boundary : list
            List of boundaries of tissue by row of downsampled image.
        """

        for i in range(row_id, limits, direction):
            left, right = boundary[i][0], boundary[i][1]
            left = left - 1 if left > 0 else 0
            right = right + 1 if (right + 1) < (n - 1) else (n - 1)
            ll, r = detect_row_boundary(i + direction, left, right)
            if ll is None:
                boundary[i + direction] = None
                break
            boundary[i + direction] = [ll, r]

    new_boundary = boundary[:]
    top, bottom = None, None
    expand_top, expand_bottom = True, True

    for i, row in enumerate(new_boundary):
        if row is None:
            continue
        if expand_bottom is False:
            new_boundary[i] = None
            continue
        if len(row) < 2:
            row.append(row[0])
        left, right = detect_row_boundary(i, row[0], row[1])
        if left is None:
            new_boundary[i] = None
            if top is None:
                expand_top = False
            if bottom is not None:
                expand_bottom = False
        else:
            new_boundary[i] = [left, right]

        if top is None:
            top = i
        bottom = i

    # detect top/bottom if necessary
    if expand_top:
        expand_row(top, 0, -1, new_boundary)

    if expand_bottom:
        expand_row(bottom, n - 1, 1, new_boundary)

    return new_boundary


def map_boundary(boundary, direction=True):
    """Map boundary to a path.

    Parameters
    ----------
    boundary : list
        List of boundaries of tissue by row of downsampled image.
    direction : bool
        Direction of path.

    Returns
    -------
    path : list
        List of boundaries of tissue by row of downsampled image.
    """
    if direction:
        start, end, step = 0, len(boundary), 1
        offset = -1
    else:
        start, end, step = len(boundary) - 1, -1, -1
        offset = 1

    def dp_shortest_path(start, end, step, offset=-1):
        """Dynamic programming shortest path.

        Parameters
        ----------
        start : int
            Starting index.
        end : int
            Ending index.

        Returns
        -------
        path : list
            List of boundaries of tissue by row of downsampled image.
        """
        dp_path = []
        dp_cost = [0, 0]
        visited = False
        for i in range(start, end, step):
            if boundary[i] is None:
                if visited:
                    break
                continue
            w = boundary[i][1] - boundary[i][0] + 1
            if not visited:
                visited = True
                dp_cost[0], dp_cost[1] = w, w
                dp_path.append([i, -1, -1])
            else:
                dp_path.append([i, 0, 0])
                for j in range(2):
                    ll = abs(boundary[i + offset][0] - boundary[i][j])
                    r = abs(boundary[i + offset][1] - boundary[i][j])
                    if ll < r:
                        dp_cost[1 - j] += w + ll
                        dp_path[-1][2 - j] = 0
                    else:
                        dp_cost[1 - j] += w + r
                        dp_path[-1][2 - j] = 1

        # reverse path
        if dp_cost[0] < dp_cost[1]:
            idx = 0
        else:
            idx = 1
        path = []
        for item in reversed(dp_path):
            x = item[0]
            if idx == 0:
                path += map(lambda y: (x, y), range(boundary[x][0], boundary[x][1] + 1))
            else:
                path += map(
                    lambda y: (x, y), range(boundary[x][1], boundary[x][0] - 1, -1)
                )
            idx = item[idx + 1]
        path.reverse()
        return path

    result = dp_shortest_path(start, end, step, offset)
    return result


def find_cell_boundary_3d(z_stack_image):
    """A default label volume image function

    Parameters
    ----------
    z_stack_image : ndarray
        A 3d array image data

    Returns
    -------
    labels : ndarray
        Labeled array
    """
    denoised_image = median_filter(z_stack_image, size=3)
    thresholded_image = denoised_image > filters.threshold_otsu(denoised_image)
    filled_images = binary_fill_holes(thresholded_image)
    cell_labels = measure.label(filled_images)

    return cell_labels


def map_labels(
    labeled_image,
    position,
    z_start,
    z_step,
    x_direction,
    y_direction,
    current_pixel_size,
    current_image_width,
    current_image_height,
    target_pixel_size,
    target_image_width,
    target_image_height,
    overlap=0.05,
    filter_pixel_number=10,
):
    """Map labels to positions

    Parameters
    ----------
    labeled_image : ndarray
        Labeled image data
    position : array[int]
        position of x, y, z, theta, f
    z_start : float
        Z start position
    z_step : float
        step of Z
    current_pixel_size : float
        Current camera pixel size
    current_image_width : int
        Current image width
    current_image_height : int
        Current image height
    target_pixel_size : float
        Target camera pixel size
    target_image_width : int
        Target image width
    target_image_height : int
        Target image height
    overlap : float
        Overlap ratio
    filter_pixel_number : int
        The pixel number to filter a label

    Returns
    -------
    z_range : int
        The maximum number of z steps
    positions : array
        Array of positions
    target_labels : array
        Array of target label index.
    """
    if target_pixel_size >= current_pixel_size:
        return 1, [position], [0]
    if overlap < 0:
        overlap = 0
    if x_direction not in ["x", "-x", "y", "-y"]:
        x_direction = "x"
    if y_direction not in ["x", "-x", "y", "-y"]:
        y_direction = "y"

    assert x_direction[-1] != y_direction[-1]

    x_direction_alignment = -1 if x_direction[0] == "-" else 1
    y_direction_alignment = -1 if y_direction[0] == "-" else 1

    x_direction_index = 0 if x_direction[-1] == "x" else 1
    y_direction_index = 1 - x_direction_index

    target_num = np.max(labeled_image)
    position_table = []
    x, y, z, theta, f = position

    center_x = current_image_width // 2
    center_y = current_image_height // 2

    x_pixel = int(target_image_width * target_pixel_size / current_pixel_size)
    y_pixel = int(target_image_height * target_pixel_size / current_pixel_size)

    regionprops = measure.regionprops(labeled_image)
    z_range = 1

    target_labels_index = []
    start = -1
    end = -1
    start_weight = current_image_width + current_image_height
    end_weight = 0

    for i in range(target_num):
        min_z, min_y, min_x, max_z, max_y, max_x = regionprops[i].bbox

        # do not need to calculate the position
        # if the label area is smaller than filter_pixel_number in x or y.
        if (max_x - min_x) < filter_pixel_number or (
            max_y - min_y
        ) < filter_pixel_number:
            continue

        target_labels_index.append(i)
        # find the start and end position
        centroid_z, centroid_y, centroid_x = regionprops[i].centroid 
        if start < 0:
            start = len(target_labels_index) - 1
            start_weight = centroid_x + centroid_y
        elif centroid_x + centroid_y < start_weight:
            if end_weight < start_weight:
                end = start
                end_weight = start_weight
            start = len(target_labels_index) - 1
            start_weight = centroid_x + centroid_y
        elif centroid_x + centroid_y > end_weight:
            end = len(target_labels_index) - 1
            end_weight = centroid_x + centroid_y

    # build a graph
    weight_graph = [[0 for i in range(len(target_labels_index))] for _ in range(len(target_labels_index))]
    for i in range(len(target_labels_index)):
        centroid_z_i, centroid_y_i, centroid_x_i = regionprops[target_labels_index[i]].centroid
        for j in range(len(target_labels_index)):
            if i == j:
                weight_graph[j][i] = weight_graph[i][j] = current_image_width + current_image_height
                continue
            centroid_z_j, centroid_y_j, centroid_x_j = regionprops[target_labels_index[j]].centroid
            weight_graph[i][j] = abs(centroid_x_i - centroid_x_j) + abs(centroid_y_i - centroid_y_j)
            weight_graph[j][i] = weight_graph[i][j]
    # find the appoximation shortest path in (x, y).
    visited_num = 1
    target_labels = [target_labels_index[start]]
    pre = start
    visited = [False] * len(target_labels_index)
    visited[start] = True
    while visited_num < len(target_labels_index):
        while True:
            idx = weight_graph[pre].index(min(weight_graph[pre]))
            weight_graph[pre][idx] = weight_graph[idx][pre] = current_image_width + current_image_height
            if not visited[idx]:
                break
        target_labels.append(target_labels_index[idx])
        pre = idx
        visited[idx] = True
        visited_num += 1
    
    for i in target_labels:
        min_z, min_y, min_x, max_z, max_y, max_x = regionprops[i].bbox

        num_x = math.ceil((max_x - min_x) / (x_pixel * (1 - overlap)))
        shift_x = (num_x * x_pixel - (max_x - min_x)) // 2

        num_y = math.ceil((max_y - min_y) / (y_pixel * (1 - overlap)))
        shift_y = (num_y * y_pixel - (max_y - min_y)) // 2

        z_range = max(z_range, (max_z - min_z))

        min_x -= shift_x
        min_y -= shift_y

        z_pos = z + z_start + min_z * z_step

        x_start = (
            position[x_direction_index]
            + x_direction_alignment
            * (min_x + x_pixel / 2 - center_x)
            * current_pixel_size
        )
        x_positions = [x_start]
        for _ in range(1, num_x):
            x_positions.append(
                x_start
                + x_direction_alignment * (x_pixel * (1 - overlap) * current_pixel_size)
            )

        y_start = (
            position[y_direction_index]
            + y_direction_alignment
            * (min_y + y_pixel / 2 - center_y)
            * current_pixel_size
        )
        y_positions = [y_start]
        for _ in range(1, num_y):
            y_positions.append(
                y_start
                + y_direction_alignment * (y_pixel * (1 - overlap) * current_pixel_size)
            )

        if x_direction_index == 0:
            position_table += [
                [x_pos, y_pos] + [z_pos, theta, f]
                for x_pos, y_pos in product(x_positions, y_positions)
            ]
        else:
            position_table += [
                [y_pos, x_pos] + [z_pos, theta, f]
                for x_pos, y_pos in product(x_positions, y_positions)
            ]

    return z_range, position_table, target_labels



# Copyright (c) 2021-2024  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Standard Imports
import logging

# Third Party Imports

# Local Imports
from navigate.model.devices.APIs.asi.asi_tiger_controller import (
    TigerController,
    ASIException,
)

p = __name__.split(".")[1]
logger = logging.getLogger(p)


class MFCTwoThousand(TigerController):
    """ASI MFC2000 Controller Class"""

    def set_speed_as_percent_max(self, pct):
        """Set speed as a percentage of the maximum speed

        Parameters
        ----------
        pct : float
            Percentage of the maximum speed
        """
        if self.default_axes_sequence is None:
            error_statement = "Default axes sequence is not set. Cannot set speed."
            logger.error(error_statement)
            raise ASIException(error_statement)

        if self._max_speeds is None:
            # First, set the speed crazy high
            self.send_command(
                f"SPEED {' '.join([f'{ax}=1000' for ax in self.default_axes_sequence])}\r"  # noqa
            )
            self.read_response()

            # Next query the maximum speed
            self.send_command(
                f"SPEED {' '.join([f'{ax}?' for ax in self.default_axes_sequence])}\r"
            )
            res = self.read_response()
            new_max_speed = float(res.split()[0].split("=")[1])
            print(f"new_max_speed: {new_max_speed}")
            self._max_speeds = [new_max_speed * 1000]

        # Now set to pct
        self.send_command(
            f"SPEED {' '.join([f'{ax}={pct*speed:.7f}' for ax, speed in zip(self.default_axes_sequence, self._max_speeds)])}\r"  # noqa
        )
        self.read_response()



# Copyright (c) 2021-2024  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only
# (subject to the limitations in the disclaimer below)
# provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Standard Imports
import tkinter as tk
from tkinter import ttk
import logging
from typing import Dict
from pathlib import Path

# Third Party Imports

# Local Imports
from navigate.view.custom_widgets.hover import HoverButton
from navigate.view.custom_widgets.validation import ValidatedSpinbox, ValidatedCombobox
from navigate.view.custom_widgets.LabelInputWidgetFactory import LabelInput
from navigate.view.custom_widgets.common import uniform_grid
import navigate

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


class ChannelsTab(tk.Frame):
    """Channels Tab for the Main Window

    This tab is used to set the channels for the stack acquisition.
    """

    def __init__(
        self,
        settings_notebook: "navigate.view.main_window_content.settings_notebook.SettingsNotebook",
        *args: list,
        **kwargs: dict,
    ):
        """Initialization of the Channels Tab

        Parameters
        ----------
        settings_notebook : SettingsNotebook
            The notebook that this tab is added to
        *args : list
            Positional arguments for tk.Frame
        **kwargs : dict
            Keyword arguments for tk.Frame

        """
        # Init Frame
        tk.Frame.__init__(self, settings_notebook, *args, **kwargs)

        #: int: The index of the tab
        self.index = 0

        #: ChannelCreator: The frame that holds the channel settings
        self.channel_widgets_frame = ChannelCreator(self)
        self.channel_widgets_frame.grid(
            row=0, column=0, columnspan=3, sticky=tk.NSEW, padx=10, pady=10
        )

        #: StackAcquisitionFrame: The frame that holds the stack acquisition settings
        self.stack_acq_frame = StackAcquisitionFrame(self)
        self.stack_acq_frame.grid(
            row=1, column=0, columnspan=3, sticky=tk.NSEW, padx=10, pady=10
        )

        #: StackTimePointFrame: The frame that holds the time settings
        self.stack_timepoint_frame = StackTimePointFrame(self)
        self.stack_timepoint_frame.grid(
            row=3, column=0, columnspan=3, sticky=tk.NSEW, padx=10, pady=10
        )

        #: MultiPointFrame: The frame that holds the multipoint settings
        self.multipoint_frame = MultiPointFrame(self)
        self.multipoint_frame.grid(
            row=4, column=0, columnspan=1, sticky=tk.NSEW, padx=10, pady=10
        )

        #: QuickLaunchFrame: The frame that holds the quick launch buttons
        self.quick_launch = QuickLaunchFrame(self)
        self.quick_launch.grid(
            row=4, column=1, columnspan=2, sticky=tk.NSEW, padx=10, pady=10
        )


class ChannelCreator(ttk.Labelframe):
    """Channel Creator

    This frame is used to create the channels for the stack acquisition.
    """

    def __init__(
        self,
        channels_tab: "navigate.view.main_window_content.settings_notebook.SettingsNotebook",
        *args: list,
        **kwargs: dict,
    ) -> None:
        """Initialization of the Channel Creator

        Parameters
        ----------
        channels_tab : tk.Frame
            The frame that this frame is added to
        *args : list
            Positional arguments for ttk.Labelframe
        **kwargs : dict
            Keyword arguments for ttk.Labelframe
        """
        #: str: The title of the frame
        self.title = "Channel Settings"
        ttk.Labelframe.__init__(self, channels_tab, text=self.title, *args, **kwargs)

        #: int: The default padding for widgets in the x direction
        self.pad_x = 1

        #: int: The default padding for widgets in the y direction
        self.pad_y = 1

        #: list: List of the variables for the channel check buttons
        self.channel_variables = []

        #: list: List of the channel check buttons
        self.channel_checks = []

        #: list: List of the variables for the laser dropdowns
        self.laser_variables = []

        #: list: List of the laser dropdowns
        self.laser_pulldowns = []

        #: list: List of the variables for the laser power dropdowns
        self.laserpower_variables = []

        #: list: List of the laser power dropdowns
        self.laserpower_pulldowns = []

        #: list: List of the variables for the filter wheel dropdowns
        self.filterwheel_variables = []

        #: list: List of the filter wheel dropdowns
        self.filterwheel_pulldowns = []

        #: list: List of the variables for the exposure time dropdowns
        self.exptime_variables = []

        #: list: List of the exposure time dropdowns
        self.exptime_pulldowns = []

        #: list: List of the variables for the time interval spin boxes
        self.interval_variables = []

        #: list: List of the time interval spin boxes
        self.interval_spins = []

        #: list: List of the variables for the defocus spin boxes
        self.defocus_variables = []

        #: list: List of the defocus spin boxes
        self.defocus_spins = []

        #: list: List of the labels for the columns
        self.label_text = [
            "Channel",
            "Laser",
            "Power",
            "Filter",
            "Exp. Time (ms)",
            "Interval",
            "Defocus",
        ]

        #: list: List of the labels for the columns
        self.labels = []

        #: list: List of the frames for the columns
        self.frame_columns = []

    def populate_frame(
        self, channels: int, filter_wheels: int, filter_wheel_names: list
    ) -> None:
        """Populates the frame with the widgets.

        This function populates the frame with the widgets for the channels. By updating
        the self.label_text list, the columns can be changed. This function is called
        when the frame is initialized, and when the number of channels is changed in the
        controller.

        Parameters
        ----------
        channels : int
            The number of channels to be added to the frame.
        filter_wheels : int
            The number of filter wheels
        filter_wheel_names : list
            The names of the filter wheels
        """

        self.create_labels(filter_wheel_names, filter_wheels)

        # Configure the columns for consistent spacing
        for i in range(len(self.label_text)):
            self.columnconfigure(i, weight=1)
        for i in range(channels):
            self.rowconfigure(i, weight=1, uniform="1")

        # Creates the widgets for each channel - populates the rows.
        for num in range(0, channels):
            self.channel_variables.append(tk.BooleanVar())
            self.channel_checks.append(
                ttk.Checkbutton(
                    self, text="CH" + str(num + 1), variable=self.channel_variables[num]
                )
            )
            self.channel_checks[num].grid(
                row=num + 1,
                column=(column_id := 0),
                sticky=tk.NSEW,
                padx=self.pad_x,
                pady=self.pad_y,
            )

            # Laser Dropdowns
            self.laser_variables.append(tk.StringVar())
            self.laser_pulldowns.append(
                ttk.Combobox(self, textvariable=self.laser_variables[num], width=6)
            )
            self.laser_pulldowns[num].config(state="readonly")
            self.laser_pulldowns[num].grid(
                row=num + 1,
                column=(column_id := column_id + 1),
                sticky=tk.NSEW,
                padx=self.pad_x,
                pady=self.pad_y,
            )

            # Laser Power Spinbox
            self.laserpower_variables.append(tk.StringVar())
            self.laserpower_pulldowns.append(
                ValidatedSpinbox(
                    self, textvariable=self.laserpower_variables[num], width=4
                )
            )
            self.laserpower_pulldowns[num].grid(
                row=num + 1,
                column=(column_id := column_id + 1),
                sticky=tk.NSEW,
                padx=self.pad_x,
                pady=self.pad_y,
            )

            # FilterWheel Dropdowns
            for i in range(filter_wheels):
                self.filterwheel_variables.append(tk.StringVar())
                self.filterwheel_pulldowns.append(
                    ttk.Combobox(
                        self, textvariable=self.filterwheel_variables[-1], width=10
                    )
                )
                self.filterwheel_pulldowns[-1].config(state="readonly")
                self.filterwheel_pulldowns[-1].grid(
                    row=num + 1,
                    column=(column_id := column_id + 1),
                    sticky=tk.NSEW,
                    padx=self.pad_x,
                    pady=self.pad_y,
                )

            # Exposure Time Spin boxes
            self.exptime_variables.append(tk.StringVar())
            self.exptime_pulldowns.append(
                ValidatedSpinbox(
                    self, textvariable=self.exptime_variables[num], width=7
                )
            )
            self.exptime_pulldowns[num].grid(
                row=num + 1,
                column=(column_id := column_id + 1),
                sticky=tk.NSEW,
                padx=self.pad_x,
                pady=self.pad_y,
            )

            # Time Interval Spin boxes
            self.interval_variables.append(tk.StringVar())
            self.interval_spins.append(
                ValidatedSpinbox(
                    self, textvariable=self.interval_variables[num], width=3
                )
            )
            self.interval_spins[num].grid(
                row=num + 1,
                column=(column_id := column_id + 1),
                sticky=tk.NSEW,
                padx=self.pad_x,
                pady=self.pad_y,
            )

            # Defocus Spinbox
            self.defocus_variables.append(tk.DoubleVar())
            self.defocus_spins.append(
                ValidatedSpinbox(
                    self, textvariable=self.defocus_variables[num], width=4
                )
            )
            self.defocus_spins[num].grid(
                row=num + 1,
                column=(column_id := column_id + 1),
                sticky=tk.NSEW,
                padx=self.pad_x,
                pady=self.pad_y,
            )

    def create_labels(self, filter_wheel_names: list, filter_wheels: int) -> None:
        """Create the labels for the columns.

        Function to create the labels for the columns of the Channel Creator frame.

        Parameters
        ----------
        filter_wheel_names : list
            A list of the names of the filter wheels
        filter_wheels : int
            Number of filter wheels
        """
        # Create the labels for the columns.
        self.label_text = [
            "Channel",
            "Laser",
            "Power",
        ]
        for i in range(filter_wheels):
            self.label_text.append(filter_wheel_names[i])

        self.label_text += ["Exp. Time (ms)", "Interval", "Defocus"]

        for idx in range(len(self.label_text)):
            self.frame_columns.append(ttk.Frame(self))
            self.frame_columns[idx].grid(
                row=0, column=idx, sticky=tk.NSEW, padx=self.pad_x, pady=self.pad_y
            )
            self.labels.append(
                ttk.Label(self.frame_columns[idx], text=self.label_text[idx])
            )
            self.labels[idx].grid(
                row=0, column=0, sticky=tk.N, pady=self.pad_y, padx=self.pad_x
            )


class StackAcquisitionFrame(ttk.Labelframe):
    """This class is the frame that holds the stack acquisition settings."""

    def __init__(self, settings_tab: ChannelsTab, *args: list, **kwargs: dict) -> None:
        """Initialization of the Stack Acquisition Frame

        Parameters
        ----------
        settings_tab : ChannelsTab
            The frame that holds the settings tab.
        *args : list
            Variable length argument list.
        **kwargs : dict
            Arbitrary keyword arguments.
        """
        # Init Frame
        text_label = "Stack Acquisition Settings (" + "\N{GREEK SMALL LETTER MU}" + "m)"
        ttk.Labelframe.__init__(self, settings_tab, text=text_label, *args, **kwargs)

        #: dict: Dictionary of the widgets in the frame
        self.inputs = {}

        #: dict: Dictionary of the buttons in the frame
        self.buttons = {}

        #: dict: Dictionary of variables in the frame
        self.additional_stack_setting_variables = {}

        self.stack_frame = ttk.Frame(self)
        self.additional_stack_frame = ttk.Frame(self)
        self.stack_frame.grid(row=0, column=0, sticky=tk.NSEW)
        self.additional_stack_frame.grid(row=1, column=0, sticky=tk.NSEW)

        self.rowconfigure(0, weight=1)
        self.rowconfigure(1, weight=1)
        self.columnconfigure(0, weight=1)

        # Start Pos Frame (Vertically oriented)
        start_names = ["start_position", "start_focus"]
        start_labels = ["Pos", "Foc"]

        #: ttk.Label: The label for the start position frame
        start_label = ttk.Label(self.stack_frame, text="Start")
        start_label.grid(row=0, column=0, sticky="S")
        for i in range(len(start_names)):
            self.inputs[start_names[i]] = LabelInput(
                parent=self.stack_frame,
                label=start_labels[i],
                input_class=ValidatedSpinbox,
                input_var=tk.DoubleVar(),
                input_args={"width": 6},
            )
            self.inputs[start_names[i]].grid(
                row=i + 1, column=0, sticky="N", pady=2, padx=(6, 0)
            )
            self.inputs[start_names[i]].label.grid(sticky="N")

        # Start button
        self.buttons["set_start"] = HoverButton(
            self.stack_frame, text="Set Start Pos/Foc"
        )
        self.buttons["set_start"].grid(row=3, column=0, sticky="N", pady=2, padx=(6, 0))

        # End Pos Frame (Vertically Oriented)
        end_names = ["end_position", "end_focus"]
        end_labels = ["Pos", "Foc"]

        #: ttk.Label: The label for the end position
        end_label = ttk.Label(self.stack_frame, text="End")
        end_label.grid(row=0, column=1, sticky="S")
        for i in range(len(end_names)):
            self.inputs[end_names[i]] = LabelInput(
                parent=self.stack_frame,
                label=end_labels[i],
                input_class=ValidatedSpinbox,
                input_var=tk.DoubleVar(),
                input_args={"width": 6},
            )
            self.inputs[end_names[i]].grid(
                row=i + 1, column=1, sticky="N", pady=2, padx=(6, 0)
            )
            self.inputs[end_names[i]].label.grid(sticky="N")

        # End Button
        self.buttons["set_end"] = HoverButton(self.stack_frame, text="Set End Pos/Foc")
        self.buttons["set_end"].grid(row=3, column=1, sticky="N", pady=2, padx=(6, 0))

        #: ttk.Label: The label for the step size
        step_size_label = ttk.Label(self.stack_frame, text="Step Size")
        step_size_label.grid(row=0, column=2, sticky="S")
        self.inputs["step_size"] = LabelInput(
            parent=self.stack_frame,
            input_class=ValidatedSpinbox,
            input_var=tk.DoubleVar(),
            input_args={"width": 6},
        )
        self.inputs["step_size"].grid(row=1, column=2, sticky="N", padx=6)

        # Slice Frame (Vertically oriented)
        #: ttk.Label: The label to add empty space to the slice frame
        self.empty_label = ttk.Label(self.stack_frame, text=" ")
        self.empty_label.grid(row=0, column=3, sticky="N")
        self.inputs["number_z_steps"] = LabelInput(
            parent=self.stack_frame,
            label="Z Slices".ljust(20),
            input_class=ValidatedSpinbox,
            input_var=tk.DoubleVar(),
            input_args={"width": 6},
        )
        self.inputs["number_z_steps"].widget.configure(state="disabled")
        self.inputs["number_z_steps"].grid(
            row=1, column=3, sticky="NSEW", pady=2, padx=(6, 0)
        )

        # devices
        self.inputs["z_device"] = LabelInput(
            parent=self.stack_frame,
            label="Z-Stack Device".ljust(30),
            input_class=ValidatedCombobox,
            input_var=tk.StringVar(),
            input_args={"width": 8},
        )
        self.inputs["z_device"].state(["disabled", "readonly"])
        self.inputs["z_device"].grid(row=4, column=0, columnspan=2, sticky="NSEW", padx=6, pady=5)

        self.inputs["f_device"] = LabelInput(
            parent=self.stack_frame,
            label="Focus Device".ljust(30),
            input_class=ValidatedCombobox,
            input_var=tk.StringVar(),
            input_args={"width": 8},
        )
        self.inputs["f_device"].state(["disabled", "readonly"])
        self.inputs["f_device"].grid(row=5, column=0, columnspan=2, sticky="NSEW", padx=6, pady=5)

        # Laser Cycling Settings
        self.inputs["cycling"] = LabelInput(
            parent=self.stack_frame,
            label="Laser Cycling Settings".ljust(25),
            input_class=ValidatedCombobox,
            input_var=tk.StringVar(),
            input_args={"width": 8},
        )
        self.inputs["cycling"].state(["readonly"])
        self.inputs["cycling"].grid(row=6, column=0, columnspan=2, sticky="NSEW", padx=6, pady=5)

        self.cubic_frame = ttk.Frame(self.stack_frame)
        self.cubic_frame.grid(row=3, rowspan=3, column=2, columnspan=2, sticky=tk.NE, padx=(5, 15), pady=(5, 0))
        
        image_directory = Path(__file__).resolve().parent

        self.image = tk.PhotoImage(
            file=image_directory.joinpath("images", "cubic_bottom_to_top.png")
        )

        # Use ttk.Label
        self.cubic_image_label = ttk.Label(self.cubic_frame, image=self.image)
        self.cubic_image_label.grid(row=0, rowspan=2, column=0, columnspan=2, sticky=tk.NSEW, padx=(5, 0), pady=(5, 0))

        self.inputs["top"] = LabelInput(
            parent=self.cubic_frame,
            label="",
            input_class=ttk.Entry,
            input_var=tk.DoubleVar(),
            input_args={"width": 6},
        )
        self.inputs["top"].grid(row=0, column=2, sticky=tk.EW, padx=0, pady=(15, 0))
        self.inputs["top"].widget.configure(state="disabled")

        self.inputs["bottom"] = LabelInput(
            parent=self.cubic_frame,
            label="",
            input_class=ttk.Entry,
            input_var=tk.DoubleVar(),
            input_args={"width": 6},
        )
        self.inputs["bottom"].grid(row=1, column=2, sticky=tk.EW, padx=0, pady=(10, 0))
        self.inputs["bottom"].widget.configure(state="disabled")

        self.inputs["z_offset"] = LabelInput(
            parent=self.additional_stack_frame,
            label="Z Offset".ljust(30),
            input_class=ValidatedSpinbox,
            input_var=tk.DoubleVar(),
            input_args={"width": 8},
        )
        self.inputs["z_offset"].widget.configure(state="disabled")
        self.inputs["z_offset"].grid(row=0, column=0, columnspan=2, sticky="NSEW", padx=6, pady=5)

        uniform_grid(self)

        # Initialize DescriptionHovers
        self.inputs["step_size"].widget.hover.setdescription("The Z-stack step size.")
        self.inputs["cycling"].widget.hover.setdescription(
            "Per Stack - Captures a full z-stack for each channel in sequence, \n"
            "scanning from the start to end Z positions for one channel before \n"
            "moving on to the next. \n\n"
            "Per Channel - Captures the full z-stack in an interleaved fashion, \n"
            "acquiring every channel at each z-plane before moving to the next."
        )
        self.inputs["z_offset"].widget.hover.setdescription(
            "The relative offset between the Z stages, if applicable."
        )
        self.buttons["set_end"].hover.setdescription(
            "Sets the Z-stack end position " "for the F and Z Axes."
        )
        self.buttons["set_start"].hover.setdescription(
            "Sets the Z-stack start position " "for the F and Z Axes."
        )
        self.inputs["z_device"].widget.hover.setdescription(
            "The device that controls the Z-stack."
        )
        self.inputs["start_position"].widget.hover.setdescription(
            "The relative starting Z position for the Z-stack."
        )
        self.inputs["start_focus"].widget.hover.setdescription(
            "The relative starting F position for the Z-stack."
        )
        self.inputs["end_position"].widget.hover.setdescription(
            "The relative ending Z position for the Z-stack."
        )
        self.inputs["end_focus"].widget.hover.setdescription(
            "The relative ending F position for the Z-stack."
        )
        self.inputs["number_z_steps"].widget.hover.setdescription(
            "The number of Z slices in the Z stack per channel."
        )
        # self.inputs["abs_z_start"].widget.hover.setdescription(
        #     "The absolute Z start position for the Z-stack."
        # )
        # self.inputs["abs_z_end"].widget.hover.setdescription(
        #     "The absolute Z end position for the Z-stack."
        # )

    def create_additional_stack_widgets(self, axes: list, devices: dict) -> None:
        """Create the additional stack widgets.

        This function creates the additional stack widgets for the
        stack acquisition settings.

        Parameters
        ----------
        axes : list
            The list of axes to create the widgets for.
        devices : dict
            The dictionary of devices to create the widgets for. {axis: device_name}
        """
        # Create the additional stack widgets here
        for widget in self.additional_stack_frame.winfo_children():
            widget.destroy()
        self.additional_stack_setting_variables = {}
        self.selected_axes_num = 0

        if len(axes) <= 2:
            return
        
        self.devices_dict = devices

        # Create the additional stack widgets here
        separator = ttk.Separator(self.additional_stack_frame, orient=tk.HORIZONTAL)
        separator.grid(row=0, column=0, columnspan=10, sticky=tk.NSEW, pady=(5, 0))

        # Stacking on axes
        label = ttk.Label(self.additional_stack_frame, text="Stacking on axes:")
        label.grid(row=1, column=0, sticky=tk.NSEW, padx=(5, 30), pady=(5, 0))

        for axis in axes:
            self.additional_stack_setting_variables[f"stack_{axis}"] = tk.BooleanVar()
            self.inputs[f"stack_{axis}"] = ttk.Checkbutton(
                self.additional_stack_frame,
                text=axis.upper(),
                command=self.update_setting_widgets(axis),
                variable=self.additional_stack_setting_variables[f"stack_{axis}"],
            )
            self.inputs[f"stack_{axis}"].grid(row=1, column=axes.index(axis) + 1, sticky=tk.NW, padx=(5, 10), pady=(5, 0))

        self.additional_stack_setting_frame = ttk.Frame(self.additional_stack_frame)
        self.additional_stack_setting_frame.grid(row=2, column=0, columnspan=10, sticky=tk.NSEW, padx=(5, 30), pady=(5, 0))
        self.additional_stack_setting_labels = {}

        for i, label_text in enumerate(["Axis", "Device", "Offset (" + "\N{GREEK SMALL LETTER MU}" + "m)"]): #, "Step", "Slice Num"]):
            label = ttk.Label(self.additional_stack_setting_frame, text=label_text)
            label.grid(row=0, column=i, sticky=tk.NSEW, padx=10, pady=2)
        for i, axis in enumerate(axes):
            label = ttk.Label(self.additional_stack_setting_frame, text=axis.upper())
            label.grid(row=i+1, column=0, sticky=tk.NSEW, padx=10, pady=2)
            label.grid_remove()
            self.additional_stack_setting_labels[axis] = label
            # Create the device label
            label = ttk.Label(self.additional_stack_setting_frame, text=self.devices_dict[axis])
            label.grid(row=i+1, column=1, sticky=tk.NSEW, padx=10, pady=2)
            label.grid_remove()
            self.additional_stack_setting_labels[f"{axis}_device"] = label
            # Create the offset spinbox
            index_name = f"{axis}_offset"
            self.additional_stack_setting_variables[index_name] = tk.DoubleVar()
            self.inputs[index_name] = ValidatedSpinbox(
                master=self.additional_stack_setting_frame,
                from_=-10000,
                to=10000,
                textvariable=self.additional_stack_setting_variables[index_name]
            )
            self.inputs[index_name].grid(row=i+1, column=2, sticky=tk.NSEW, padx=10, pady=2)
            self.inputs[index_name].grid_remove()

        self.additional_stack_setting_frame.grid_remove()

        uniform_grid(self.additional_stack_frame)

    def update_setting_widgets(self, axis: str) -> None:
        """Update the setting widgets based on the movement mode.

        This function updates the setting widgets based on the
        movement mode selected.

        Parameters
        ----------
        axis : str
            The axis to update the widgets for.
        """
        def func(*args: list) -> None:
            """Inner function to update the widgets.

            Parameters
            ----------
            *args : list
                The arguments passed to the function.
            """
            if self.additional_stack_setting_variables[f"stack_{axis}"].get():
                self.additional_stack_setting_frame.grid()
                self.inputs[f"{axis}_offset"].grid()
                self.additional_stack_setting_labels[axis].grid()
                self.additional_stack_setting_labels[f"{axis}_device"].grid()
                self.selected_axes_num += 1
            else:
                self.inputs[f"{axis}_offset"].grid_remove()
                self.additional_stack_setting_labels[axis].grid_remove()
                self.additional_stack_setting_labels[f"{axis}_device"].grid_remove()
                self.selected_axes_num -= 1
                if self.selected_axes_num <= 0:
                    self.additional_stack_setting_frame.grid_remove()

        return func
    
    # Getters
    def get_variables(self) -> dict:
        """Returns a dictionary of the variables in the widget

        This function returns a dictionary of all the variables
        that are tied to each widget name.

        The key is the widget name, value is the variable associated.

        Returns
        -------
        dict
            Dictionary of the variables in the widget
        """
        variables = {}
        for key, widget in self.inputs.items():
            variables[key] = widget.get_variable()
        return variables

    def get_widgets(self) -> dict:
        """Returns a dictionary of the widgets.

        This function returns the dictionary that holds the input widgets.
        The key is the widget name, value is the LabelInput class that has all the data.

        Returns
        -------
        dict
            Dictionary of the widgets in the widget
        """
        return self.inputs

    def get_buttons(self) -> dict:
        """Returns a dictionary of the buttons in the frame.

        This function returns the dictionary that holds the buttons.
        The key is the button name, value is the button.

        Returns
        -------
        dict
            Dictionary of the buttons in the widget
        """
        return self.buttons


class StackTimePointFrame(ttk.Labelframe):
    """Frame for the stack time point settings in the channels tab.

    This class is a frame that holds the widgets for the stack time point settings.
    It is a subclass of ttk.Labelframe.
    """

    def __init__(self, settings_tab: ChannelsTab, *args: list, **kwargs: dict) -> None:
        """Initialization of the Stack Time point Frame

        Parameters
        ----------
        settings_tab : ChannelsTab
            The frame that the stack time point frame will be placed in
        *args : list
            Variable length argument list
        **kwargs : dict
            Arbitrary keyword arguments
        """
        text_label = "Timepoint Settings"
        ttk.Labelframe.__init__(self, settings_tab, text=text_label, *args, **kwargs)

        #: dict: Dictionary of the widgets in the frame
        self.inputs = {}

        #: ttk.Label: The label for the save data checkbox
        self.laser_label = ttk.Label(self, text="Save Data")
        self.laser_label.grid(row=0, column=0, sticky=tk.NSEW, padx=(4, 5), pady=(4, 0))

        #: tk.BooleanVar: The variable for the save data checkbox
        self.save_data = tk.BooleanVar()
        self.save_data.set(False)

        #: ttk.Checkbutton: The save data checkbox
        self.save_check = ttk.Checkbutton(self, text="", variable=self.save_data)
        self.save_check.grid(row=0, column=1, sticky=tk.NSEW, pady=(4, 0))
        self.inputs["save_check"] = self.save_check

        #: ttk.Label: The label for the timepoints spinbox
        self.filterwheel_label = ttk.Label(self, text="Timepoints")
        self.filterwheel_label.grid(
            row=1, column=0, sticky=tk.NSEW, padx=(4, 5), pady=2
        )

        #: tk.StringVar: The variable for the timepoints spinbox
        self.exp_time_spinval = tk.StringVar()

        #: ValidatedSpinbox: The timepoints spinbox
        self.exp_time_spinbox = ValidatedSpinbox(
            self, textvariable=self.exp_time_spinval, width=3
        )
        self.exp_time_spinbox.grid(row=1, column=1, sticky=tk.NSEW, pady=2)
        self.inputs["time_spin"] = self.exp_time_spinbox

        #: ttk.Label: The label for the stack acquisition time spinbox
        self.exp_time_label = ttk.Label(self, text="Stack Acq. Time")
        self.exp_time_label.grid(row=2, column=0, sticky=tk.NSEW, padx=(4, 5), pady=2)

        #: tk.StringVar: The variable for the stack acquisition time spinbox
        self.stack_acq_spinval = tk.StringVar()

        #: ttk.Spinbox: The stack acquisition time spinbox
        self.stack_acq_spinbox = ttk.Spinbox(
            self, textvariable=self.stack_acq_spinval, width=6
        )
        self.stack_acq_spinbox.grid(row=2, column=1, sticky=tk.NSEW, pady=2)
        self.stack_acq_spinbox.state(["disabled"])

        #: ttk.Label: The label for the stack pause spinbox
        self.exp_time_label = ttk.Label(self, text="Stack Pause (s)")
        self.exp_time_label.grid(row=0, column=2, sticky=tk.NSEW, padx=(4, 5), pady=2)

        #: tk.StringVar: The variable for the stack pause spinbox
        self.stack_pause_spinval = tk.StringVar()

        #: ValidatedSpinbox: The stack pause spinbox
        self.stack_pause_spinbox = ValidatedSpinbox(
            self, textvariable=self.stack_pause_spinval, width=6
        )
        self.stack_pause_spinbox.grid(row=0, column=3, sticky=tk.NSEW, pady=2)
        self.inputs["stack_pause"] = self.stack_pause_spinbox

        #: ttk.Label: The label for the time point interval spinbox
        self.exp_time_label = ttk.Label(self, text="Time Interval (hh:mm:ss)")
        self.exp_time_label.grid(row=1, column=2, sticky=tk.NSEW, padx=(4, 5), pady=2)

        #: tk.StringVar: The variable for the time point interval spinbox
        self.timepoint_interval_spinval = tk.StringVar()
        if self.timepoint_interval_spinval.get() == "":
            self.timepoint_interval_spinval.set("0")

        #: ttk.Spinbox: The time point interval spinbox
        self.timepoint_interval_spinbox = ttk.Spinbox(
            self, textvariable=self.timepoint_interval_spinval, width=6
        )
        self.timepoint_interval_spinbox.grid(row=1, column=3, sticky=tk.NSEW, pady=2)
        self.timepoint_interval_spinbox.state(["disabled"])  # Starts it disabled

        #: ttk.Label: The label for the total time spinbox
        self.exp_time_label = ttk.Label(self, text="Experiment Duration (hh:mm:ss)")
        self.exp_time_label.grid(
            row=2, column=2, sticky=tk.NSEW, padx=(4, 5), pady=(2, 6)
        )

        #: tk.StringVar: The variable for the total time spinbox
        self.total_time_spinval = tk.StringVar()
        if self.total_time_spinval.get() == "":
            self.total_time_spinval.set("0")

        #: ttk.Spinbox: The total time spinbox
        self.total_time_spinval = ttk.Spinbox(
            self, textvariable=self.total_time_spinval, width=6
        )
        self.total_time_spinval.grid(row=2, column=3, sticky=tk.NSEW, pady=(2, 6))
        self.total_time_spinval.state(["disabled"])

    def get_variables(self) -> dict:
        """Returns a dictionary of all the variables that are tied to each widget name.

        The key is the widget name, value is the variable associated.

        Returns
        -------
        variables : dict
            A dictionary of all the variables that are tied to each widget name.
        """
        variables = {}
        for key, widget in self.inputs.items():
            variables[key] = widget.get_variable()
        return variables

    def get_widgets(self) -> dict:
        """Returns a dictionary of all the widgets that are tied to each widget name.

        The key is the widget name, value is the LabelInput class that has all the data.

        Returns
        -------
        widgets : dict
            A dictionary of all the widgets that are tied to each widget name.
        """
        return self.inputs


class MultiPointFrame(ttk.Labelframe):
    """Multi-Position Acquisition Frame"""

    def __init__(self, settings_tab: ChannelsTab, *args: list, **kwargs: dict) -> None:
        """Initialization of the Multi-Position Acquisition Frame

        Parameters
        ----------
        settings_tab : ChannelsTab
            The frame that the multipoint frame will be placed in
        *args : list
            Variable length argument list
        **kwargs : dict
            Arbitrary keyword arguments
        """
        text_label = "Multi-Position Acquisition"
        ttk.Labelframe.__init__(self, settings_tab, text=text_label, *args, **kwargs)

        #: ttk.Label: The label for the save data checkbox
        self.laser_label = ttk.Label(self, text="Enable")
        self.laser_label.grid(row=0, column=0, sticky=tk.NSEW, padx=(4, 4), pady=(4, 4))

        #: tk.BooleanVar: The variable for the save data checkbox
        self.on_off = tk.BooleanVar()

        #: ttk.Checkbutton: The save data checkbox
        self.save_check = ttk.Checkbutton(self, text="", variable=self.on_off)
        self.save_check.grid(row=0, column=1, sticky=tk.NSEW, pady=(4, 4))

        #: dict: Dictionary of the buttons in the frame
        self.buttons = {"tiling": ttk.Button(self, text="Launch Tiling Wizard")}
        self.buttons["tiling"].grid(
            row=0, column=2, sticky=tk.NSEW, padx=(10, 0), pady=(4, 4)
        )


class QuickLaunchFrame(ttk.Labelframe):
    """Quick Launch Buttons Frame

    This frame contains buttons that launch the Tiling Wizard.
    """

    def __init__(self, settings_tab: ChannelsTab, *args: list, **kwargs: dict) -> None:
        """Initialization of the Quick Launch Buttons Frame

        Parameters
        ----------
        settings_tab : ChannelsTab
            The settings tab object that this frame is being added to.
        *args : list
            Variable length argument list.
        **kwargs : dict
            Arbitrary keyword arguments.
        """
        text_label = "Quick Launch Buttons"
        ttk.Labelframe.__init__(self, settings_tab, text=text_label, *args, **kwargs)

        #: Dict[str, ttk.Button]: Dictionary of the buttons in the frame
        self.buttons: Dict[str, ttk.Button] = {}
        self.buttons = {
            "waveform_parameters": ttk.Button(self, text="Waveform Parameters")
        }
        self.buttons["waveform_parameters"].grid(
            row=0, column=2, sticky=tk.NSEW, padx=(4, 4), pady=(4, 4)
        )
        self.buttons["autofocus_button"] = ttk.Button(self, text="Autofocus Settings")
        self.buttons["autofocus_button"].grid(
            row=1, column=2, sticky=tk.NSEW, padx=(4, 4), pady=(4, 4)
        )



import tkinter as tk

class CollapsibleFrame(tk.Frame):
    def __init__(self, parent, title="", *args, **kwargs):
        tk.Frame.__init__(self, parent, *args, **kwargs)
        
        self.title = title
        self.visible = False
        
        # Create a label to act as a title/header
        self.label = tk.Label(self, text=self.title, bg="lightgrey", relief="raised", padx=5)
        self.label.grid(row=0, column=0, sticky=tk.NSEW)
        
        # Create a frame to hold the contents of the collapsible frame
        self.content_frame = tk.Frame(self)
        self.toggle_visibility()

        self.label.bind("<Button-1>", lambda event: self.toggle_visibility())
        
    def toggle_visibility(self):
        if self.visible:
            self.label["text"] = self.title + " " + "\u25BC"
            self.content_frame.grid_forget()  # Hide the content frame
            self.visible = False
        else:
            self.label["text"] = self.title + " " + "\u25B2"
            self.content_frame.grid(row=1, column=0, sticky=tk.NSEW)  # Show the content frame
            self.visible = True

    def fold(self):
        if self.visible:
            self.toggle_visibility()




# Copyright (c) 2021-2024  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

# Standard Library Imports
import tkinter as tk
from multiprocessing.managers import ListProxy, DictProxy
import logging
from typing import Dict, Optional, Callable, Iterable
import re

# Third Party Imports

# Local Imports
import navigate
from navigate.controller.configuration_controller import ConfigurationController
from navigate.controller.sub_controllers.gui import GUIController
from navigate.tools.decorators import log_initialization
from navigate.view.main_application_window import MainApp
from navigate.view.main_window_content.stage_tab import StageControlTab, OtherAxisFrame

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


@log_initialization
class StageController(GUIController):
    """StageController

    This class is the controller for the stage GUI. It handles the stage movement
    and the stage limits. It also handles the stage movement buttons and the
    stage movement limits.
    """

    def __init__(
        self,
        view: StageControlTab,
        main_view: MainApp,
        canvas: tk.Canvas,
        parent_controller: "navigate.controller.controller.Controller",
    ) -> None:
        """Initializes the StageController

        Parameters
         ----------
         view : navigate.view.stage_view.StageView
             The stage view
         main_view : tkinter.Tk
             The main view of the microscope
         canvas : tkinter.Canvas
             The canvas of the microscope
         parent_controller : navigate.controller.Controller
             The parent controller of the stage controller
        """
        super().__init__(view, parent_controller)

        #: dict: The joystick axes
        self.new_joystick_axes = None

        #: str: The default microscope
        self.default_microscope = (
            f"{self.parent_controller.configuration_controller.microscope_name}"
        )

        #: bool: The joystick mode is on.
        self.joystick_is_on = False

        #: list: The joystick axes
        self.joystick_axes = self.parent_controller.configuration["configuration"][
            "microscopes"
        ][self.default_microscope]["stage"].get("joystick_axes", [])

        #: dict: The stage setting dictionary
        self.stage_setting_dict = self.parent_controller.configuration["experiment"][
            "StageParameters"
        ]
        self.stage_axes = self.parent_controller.configuration_controller.stage_axes
        all_stage_axes = self.parent_controller.configuration_controller.all_stage_axes
        # add stage control buttons to the stage control tab
        for axis in all_stage_axes:
            if axis in ["x", "y", "z", "theta", "f"]:
                continue
            view.add_additional_stage(axis)
        #: dict: The event id
        self.event_id = dict(zip(all_stage_axes, [None] * len(all_stage_axes)))

        #: dict: The minimum stage positions.
        self.position_min = {}

        #: dict: The maximum stage positions.
        self.position_max = {}

        #: dict: The widget values for the stage position and step sizes.
        self.widget_vals = self.view.get_variables()

        # gui event bind
        buttons = self.view.get_buttons()
        for k in buttons:
            temp = k.split("_")
            if len(temp) < 3 or temp[-1] != "btn":
                continue
            axis = temp[-2]
            large_step = True if "large" in k else False
            if "up" == temp[-3]:
                buttons[k].configure(
                    command=self.up_btn_handler(axis=axis, large_step=large_step),
                    repeatinterval=300,
                    repeatdelay=300,
                )

            elif "down" == temp[-3]:
                buttons[k].configure(
                    command=self.down_btn_handler(axis=axis, large_step=large_step),
                    repeatinterval=300,
                    repeatdelay=300,
                )

        for k in all_stage_axes:
            if k == "x" or k == "y":
                k = "xy"
            self.widget_vals[k + "_step"].trace_add(
                "write", self.update_step_size_handler(k)
            )

        buttons["stop"].configure(command=self.stop_button_handler)
        buttons["joystick"].configure(
            command=lambda: self.view.after(250, self.joystick_button_handler)
        )

        #: dict: The position callback traces
        self.position_callback_traces = {}

        #: bool: The position callbacks are bound
        self.position_callbacks_bound = False
        self.bind_position_callbacks()

        #: bool: The stage limits are on
        self.stage_limits = True

        #: list: The stage flip flags
        self.flip_flags = None
        self.initialize()
        self.set_hover_descriptions()

        # Bind the buttons for the z-stack start and stop.
        self.view.stack_shortcuts.set_start_button.configure(
            command=self.parent_controller.channels_tab_controller.update_start_position
        )
        self.view.stack_shortcuts.set_end_button.configure(
            command=self.parent_controller.channels_tab_controller.update_end_position
        )



    def stage_key_press(self, event: tk.Event) -> None:
        """The stage key press.

        This function is called when a W, A, S, or D is
        pressed and initializes a stage movement..

        Parameters
        ----------
        event : tk.Event
            The tkinter event

        """
        if event.state != 0:
            return
        if not self.joystick_is_on:
            char = event.char.lower()
            if char == "w":
                self.up_btn_handler(axis="y", large_step=False)()
            elif char == "a":
                self.down_btn_handler(axis="x", large_step=False)()
            elif char == "s":
                self.down_btn_handler(axis="y", large_step=False)()
            elif char == "d":
                self.up_btn_handler(axis="x", large_step=False)()

    def initialize(self) -> None:
        """Initialize the Stage limits of steps and positions."""
        config = self.parent_controller.configuration_controller
        self.stage_axes = config.stage_axes
        # disable stages not available for the current microscope
        for axis in set(config.all_stage_axes) - set(self.stage_axes):
            stage_frame = getattr(self.view, f"{axis}_frame")
            stage_frame.toggle_button_states(True, [axis])
            stage_frame.increment_box.widget.config(state="disabled")
            self.view.position_frame.inputs[axis].widget.config(state="disabled")
        self.disable_synthetic_stages(config)

        # Get the minimum and maximum stage limits.
        self.position_min = config.get_stage_position_limits("_min")
        self.position_max = config.get_stage_position_limits("_max")

        widgets = self.view.get_widgets()
        step_dict = self.stage_setting_dict[config.microscope_name]
        for axis in self.stage_axes:
            # Set Stage Limits
            widgets[axis].widget.min = self.position_min[axis]
            widgets[axis].widget.max = self.position_max[axis]
            if axis == "x" or axis == "y":
                step_axis = "xy"
            else:
                step_axis = axis

            # Set step size.
            # the minimum step should be non-zero and non-negative.
            # TODO: Move to the GUI configuration file.
            widgets[step_axis + "_step"].widget.configure(from_=0.01)
            widgets[step_axis + "_step"].widget.configure(to=self.position_max[axis])
            step_increment = step_dict.get(f"{step_axis}_step", 10) // 10
            if step_increment == 0:
                step_increment = 1
            widgets[step_axis + "_step"].widget.configure(increment=step_increment)
            widgets[step_axis + "_step"].set(step_dict.get(f"{step_axis}_step", 10))

        # Joystick
        microscope_name = config.microscope_name

        self.new_joystick_axes = self.parent_controller.configuration["configuration"][
            "microscopes"
        ][f"{microscope_name}"]["stage"].get("joystick_axes", [])

        if self.view.stop_frame.joystick_btn.winfo_ismapped():
            if self.new_joystick_axes is None or list(self.new_joystick_axes) == []:
                self.view.stop_frame.joystick_btn.grid_forget()
            else:
                self.view.stop_frame.joystick_btn.grid()
        else:
            if (
                self.new_joystick_axes is not None
                and list(self.new_joystick_axes) != []
            ):
                self.view.stop_frame.joystick_btn.grid()
            else:
                self.view.stop_frame.joystick_btn.grid_forget()

        if list(self.joystick_axes) != list(self.new_joystick_axes):
            self.force_enable_all_axes()
            self.joystick_is_on = False

        self.joystick_axes = self.new_joystick_axes
        self.flip_flags = config.stage_flip_flags

    def disable_synthetic_stages(self, config: ConfigurationController) -> None:
        """Disable synthetic stages.

        Parameters
        ----------
        config : ConfigurationController
            The configuration controller
        """
        microscope_configuration = config.get_microscope_configuration_dict()
        stages = microscope_configuration["stage"]["hardware"]

        if type(stages) is ListProxy:
            stages = list(stages)
        elif type(stages) is DictProxy:
            stages = [dict(stages)]

        for stage in stages:
            if type(stage) is DictProxy:
                stage_dict = dict(stage)
            elif type(stage) is dict:
                stage_dict = stage

            for axis in stage_dict["axes"]:
                if "synthetic" in stage_dict["type"].lower():
                    state = "disabled"
                    flag = True
                else:
                    state = "normal"
                    flag = False

                if axis == "x":
                    self.view.xy_frame.toggle_button_states(flag, ["x"])
                    self.view.xy_frame.increment_box.widget.config(state=state)
                    self.view.position_frame.inputs["x"].widget.config(state=state)

                elif axis == "y":
                    self.view.xy_frame.toggle_button_states(flag, ["y"])
                    self.view.position_frame.inputs["y"].widget.config(state=state)
                
                else:
                    stage_frame = getattr(self.view, f"{axis}_frame")
                    stage_frame.toggle_button_states(flag, [axis])
                    stage_frame.increment_box.widget.config(state=state)
                    self.view.position_frame.inputs[axis].widget.config(state=state)

    def bind_position_callbacks(self) -> None:
        """Binds position_callback() to each axis, records the trace name so we can
        unbind later."""
        widgets = self.view.get_widgets()
        if not self.position_callbacks_bound:
            for axis in self.stage_axes:
                widgets[axis].widget.bind("<FocusOut>", self.position_callback(axis))
            self.position_callbacks_bound = True

    def unbind_position_callbacks(self) -> None:
        """Unbinds position callbacks."""
        if self.position_callbacks_bound:
            for axis, cbname in self.position_callback_traces.items():
                self.widget_vals[axis].trace_remove("write", cbname)
            self.position_callbacks_bound = False

    def populate_experiment_values(self) -> None:
        """This function sets all the position and step values"""
        self.stage_setting_dict = self.parent_controller.configuration["experiment"][
            "StageParameters"
        ]
        widgets = self.view.get_widgets()
        for k in widgets:
            self.widget_vals[k].set(self.stage_setting_dict.get(k, 0))
            widgets[k].widget.trigger_focusout_validation()

    def set_position(self, position: Dict[str, float]) -> None:
        """This function is to populate(set) position in the View

        Parameters
        ----------
        position : Dict[str, float]
            {'x': value, 'y': value, 'z': value, 'theta': value, 'f': value}
        """
        for axis in self.stage_axes:
            if axis not in position:
                continue
            self.widget_vals[axis].set(position[axis])
            self.position_callback(axis)()
        self.show_verbose_info("Set stage position")

    def set_position_silent(self, position: Dict[str, float]) -> None:
        """This function is to populate(set) position in the View without a trace.

        Parameters
        ----------
        position : Dict[str, float]
            {'x': value, 'y': value, 'z': value, 'theta': value, 'f': value}
        """
        widgets = self.view.get_widgets()
        for axis in self.stage_axes:
            if axis not in position:
                continue
            self.widget_vals[axis].set(position[axis])
            # validate position value if set through variable
            if self.stage_limits:
                widgets[axis].widget.trigger_focusout_validation()
            self.stage_setting_dict[axis] = position.get(axis, 0)
        self.show_verbose_info("Set stage position")

    def get_position(self) -> Optional[dict[str, float]]:
        """This function returns current position from the view.

        Returns
        -------
        position : Optional[dict[str, float]]
            Dictionary of x, y, z, theta, and f values. None if the value is not valid.

        """
        position = {}
        try:
            for axis in self.stage_axes:
                position[axis] = float(self.widget_vals[axis].get())
                if self.stage_limits is True:
                    if (
                        position[axis] < self.position_min[axis]
                        or position[axis] > self.position_max[axis]
                    ):
                        return None
        except tk.TclError:
            # Tkinter will raise error when the variable is DoubleVar and the value
            # is empty
            return None
        return position

    def up_btn_handler(
        self,
        axis: str,
        large_step: bool = False,
    ) -> Callable[[], None]:
        """This function generates command functions according to the desired axis
        to move.

        Parameters
        ----------
        axis : str
            Should be one of 'x', 'y', 'z', 'theta', 'f'
            position_axis += step_axis
        large_step : bool
            If True, the step size is 5x the normal step size.

        Returns
        -------
        handler : Callable[[], None]
            Function for setting desired stage positions in the View.
        """
        step_multiple = 5 if large_step else 1
        position_val = self.widget_vals[axis]
        if axis == "x" or axis == "y":
            step_val = self.widget_vals["xy_step"]
        else:
            step_val = self.widget_vals[axis + "_step"]

        def handler():
            """This function generates command functions according to the desired axis
            to move."""
            stage_direction = -1 if self.flip_flags[axis] else 1
            try:
                temp = (
                    float(position_val.get())
                    + step_val.get() * stage_direction * step_multiple
                )
            except (tk.TclError, ValueError):
                return
            if self.stage_limits is True:
                if temp > self.position_max[axis]:
                    temp = self.position_max[axis]
                elif temp < self.position_min[axis]:
                    temp = self.position_min[axis]
            # guarantee stage won't move out of limits
            if float(position_val.get()) != temp:
                position_val.set(temp)
                self.position_callback(axis)()

        return handler

    def down_btn_handler(
        self, axis: str, large_step: bool = False
    ) -> Callable[[], None]:
        """This function generates command functions according to the desired axis
        to move.

        Parameters
        ----------
        axis : str
            Should be one of 'x', 'y', 'z', 'theta', 'f'
            position_axis += step_axis
        large_step : bool
            If True, the step size is 5x the normal step size.

        Returns
        -------
        handler : Callable[[], None]
            Function for setting desired stage positions in the View.
        """
        step_multiple = 5 if large_step else 1
        position_val = self.widget_vals[axis]
        if axis == "x" or axis == "y":
            step_val = self.widget_vals["xy_step"]
        else:
            step_val = self.widget_vals[axis + "_step"]

        def handler():
            """This function generates command functions according to the desired axis
            to move."""
            stage_direction = -1 if self.flip_flags[axis] else 1
            try:
                temp = (
                    float(position_val.get())
                    - step_val.get() * stage_direction * step_multiple
                )
            except (tk.TclError, ValueError):
                return
            if self.stage_limits is True:
                if temp < self.position_min[axis]:
                    temp = self.position_min[axis]
                elif temp > self.position_max[axis]:
                    temp = self.position_max[axis]
            # guarantee stage won't move out of limits
            if float(position_val.get()) != temp:
                position_val.set(temp)
                self.position_callback(axis)()

        return handler

    def stop_button_handler(self, *args: Iterable) -> None:
        """This function stops the stage after a 250 ms debouncing period of time.

        Parameters
        ----------
        *args : Iterable
            Variable length argument list
        """
        self.view.after(250, lambda *args: self.parent_controller.execute("stop_stage"))

    def joystick_button_handler(
        self, event: Optional[tk.Event] = None, *args: Iterable
    ) -> None:
        """Toggle the joystick operation mode.

        Parameters
        ----------
        event : Optional[tk.Event]
            The tkinter event
        *args : Iterable
            Variable length argument list
        """
        self.joystick_is_on = not self.joystick_is_on
        self.parent_controller.execute("stop_stage")
        self.view.toggle_button_states(self.joystick_is_on, self.joystick_axes)

    def force_enable_all_axes(
        self, event: Optional[tk.Event] = None, *args: Iterable
    ) -> None:
        """Enables all buttons and entries on the stage tab.

        Parameters
        ----------
        event : Optional[tk.Event]
            The tkinter event (currently unused)
        *args : Iterable
            Variable length argument list
        """
        self.view.force_enable_all_axes()

    def position_callback(
        self, axis: str, *args: Iterable, **kwargs: dict
    ) -> Callable[[], None]:
        """Callback functions bind to position variables.

        Implements debounce functionality for user inputs (or click buttons) to reduce
        time costs of moving stage.

        Parameters
        ----------
        axis : str
            axis can be 'x', 'y', 'z', 'theta', 'f'
        *args : Iterable
            Variable length argument list
        kwargs : Iterable
            Variable length keyword argument list

        Returns
        -------
        handler : Callable[[], None]
            Function for moving stage to the desired position with debounce
            functionality.
        """
        position_var = self.widget_vals[axis]
        temp = self.view.get_widgets()
        widget = temp[axis].widget

        def handler(*_):
            """Callback functions bind to position variables."""
            # check if focus on another window
            if not self.view.focus_get():
                return
            if self.event_id[axis]:
                self.view.after_cancel(self.event_id[axis])
            # if position is not a number, then do not move stage
            try:
                position = float(position_var.get())
                if self.stage_limits:
                    widget.trigger_focusout_validation()
                    # if position is not inside limits do not move stage
                    if position < float(self.position_min[axis]) or position > float(
                        self.position_max[axis]
                    ):
                        return
            except tk.TclError:
                if self.event_id[axis]:
                    self.view.after_cancel(self.event_id[axis])
                return
            except AttributeError:
                logger.error(f"Attribute Error Caught: trying to set position {axis}")
                return

            # update stage position
            self.stage_setting_dict[axis] = position
            # Debouncing wait duration - Duration of time to integrate the number of
            # clicks that a user provides. If 1000 ms, if user hits button 10x within
            # 1s, only moves to the final value.
            self.event_id[axis] = self.view.after(
                300,
                lambda *args: self.parent_controller.execute("stage", position, axis),
            )

            self.show_verbose_info("Stage position changed")

        return handler

    def update_step_size_handler(self, axis: str) -> Callable[[], None]:
        """Callback functions bind to step size variables

        Parameters
        ----------
        axis : str
            axis can be 'xy', 'z', 'theta', 'f'

        Returns
        -------
        handler : Callable[[], None]
            Function to update step size in experiment.yml.
        """
        def func(*args):
            """Callback functions bind to step size variables."""
            microscope_name = self.parent_controller.configuration["experiment"][
                "MicroscopeState"
            ]["microscope_name"]
            try:
                step_size = int(self.widget_vals[axis + "_step"].get())
            except (ValueError, tk.TclError):
                return
            self.stage_setting_dict[microscope_name][axis + "_step"] = step_size
            # update hover descriptions
            self.set_hover_descriptions()

        return func

    def set_hover_descriptions(self) -> None:
        """Set hover descriptions for the stage tab"""

        btn_prefix = ["large_up", "large_down", "up", "down"]
        step_multiple = [5, -5, 1, -1]

        for axis in self.stage_axes:
            if axis == "x" or axis == "y":
                frame_prefix = "xy"
            else:
                frame_prefix = axis
            step_value = self.widget_vals[f"{frame_prefix}_step"].get()
            if frame_prefix == "xy":
                btn_suffix = f"{axis}_btn"
            else:
                btn_suffix = "btn"

            if axis == "theta":
                description = f"\N{DEGREE SIGN} in \N{GREEK CAPITAL LETTER THETA}."
            else:
                description = f"\N{GREEK SMALL LETTER MU}m in {axis.upper()}."

            for i in range(len(btn_prefix)):
                exec(f"self.view.{frame_prefix}_frame.{btn_prefix[i]}_{btn_suffix}.hover."
                     f"setdescription('Move {step_multiple[i] * step_value} {description}')")

        # Position Frame
        for axis in self.stage_axes:
            if axis == "theta":
                description = "Theta stage position in degrees."
            else:
                description = f"{axis.upper()} stage position in \N{GREEK SMALL LETTER MU}m."
            self.view.position_frame.inputs[axis].widget.hover.setdescription(
                description
            )

        self.view.stack_shortcuts.set_start_button.hover.setdescription(
            "Sets the start positions for Z and F for a Z-Stack."
        )
        self.view.stack_shortcuts.set_end_button.hover.setdescription(
            "Sets the end positions for Z and F for a Z-Stack."
        )

        # Stop button.
        self.view.stop_frame.joystick_btn.hover.setdescription(
            "Enables/Disables joystick mode."
        )
